# -*- coding: utf-8 -*-
"""
Created on Wed Aug 21 13:47:43 2019

@author: gbuster
"""
from concurrent.futures import as_completed
import json
import logging
import numpy as np
import os
import pandas as pd
from scipy.spatial import cKDTree
from warnings import warn

from rex.rechunk_h5 import to_records_array
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.utilities import parse_table

from reVX.handlers.outputs import Outputs
from reVX.plexos.utilities import (DataCleaner, get_coord_labels,
                                   parse_table_name)

logger = logging.getLogger(__name__)


class PlexosNode:
    """Framework to analyze the gen profile at a single plexos node."""

    def __init__(self, sc_build, cf_fpath, cf_res_gids=None,
                 forecast_fpath=None, forecast_map=None):
        """
        Parameters
        ----------
        sc_build : pd.DataFrame
            Supply curve buildout table. Must only have rows that are built
            in this plexos node. Must have resource_gid lookup, counts per
            resource_gid, and capacity at each SC point.
        cf_fpath : str
            File path to capacity factor file to get profiles from.
        cf_res_gids : list | np.ndarray, optional
            Resource GID's available in cf_fpath, if None pull from cf_fpath,
            by default None
        forecast_fpath : str | None, optional
            Forecasted capacity factor .h5 file path (reV results).
            If not None, the generation profiles are sourced from this file,
            by default None
        forecast_map : np.ndarray | None, optional
            (n, 1) array of forecast meta data indices mapped to the generation
            meta indices where n is the number of generation points. None if
            no forecast data being considered,
            by default None
        """
        self._sc_build = \
            DataCleaner.rename_cols(sc_build,
                                    name_map=DataCleaner.REV_NAME_MAP)
        self._cf_fpath = cf_fpath
        if cf_res_gids is None:
            cf_res_gids = self._get_cf_res_gids(cf_fpath)

        self._cf_res_gids = cf_res_gids
        self._forecast_fpath = forecast_fpath
        self._forecast_map = forecast_map

    @staticmethod
    def _get_cf_res_gids(cf_fpath):
        """
        Get available res_gids from cf .h5 file

        Parameters
        ----------
        cf_fpath : str
            File path to capacity factor file to get profiles from.

        Returns
        -------
        cf_res_gids : ndarray
            Array of resource GIDs available in the cf file
        """
        with Outputs(cf_fpath, mode='r') as cf_outs:
            cf_res_gids = cf_outs.get_meta_arr('gid')

        if not isinstance(cf_res_gids, np.ndarray):
            cf_res_gids = np.array(list(cf_res_gids))

        return cf_res_gids

    def _get_sc_meta(self, row_idx):
        """Get meta for SC point row index, which is part of this plexos node.

        Parameters
        ----------
        row_idx : int
            Index value for the row of the target SC point in self._sc_build.

        Returns
        -------
        sc_meta : pd.DataFrame
            Dataframe with rows corresponding to resource/generation pixels
            that are part of this SC point. Sorted by cf_mean with best
            cf_mean at top.
        buildout : float
            Total REEDS requested buildout associated with SC point i.
        """
        res_gids, gid_counts, buildout, capacity = \
            self._parse_sc_point(row_idx)

        gid_capacity = gid_counts / np.sum(gid_counts) * capacity
        gen_gids = [np.where(self._cf_res_gids == g)[0][0]
                    for g in res_gids]
        sc_meta = pd.DataFrame({'gen_gid': gen_gids,
                                'res_gid': res_gids,
                                'gid_count': gid_counts,
                                'gid_capacity': gid_capacity})
        sc_meta = sc_meta.sort_values(by='gen_gid')

        with Outputs(self._cf_fpath, mode='r') as cf_outs:
            cf_mean = cf_outs['cf_mean', list(sc_meta['gen_gid'].values)]

        sc_meta['cf_mean'] = cf_mean
        sc_meta = sc_meta.sort_values(by='cf_mean', ascending=False)
        sc_meta = sc_meta.reset_index(drop=True)

        # infinite capacity in the last gid to make sure full buildout is done
        sc_meta.loc[sc_meta.index[-1], 'gid_capacity'] = 1e6

        return sc_meta, buildout

    def _parse_sc_point(self, row_idx):
        """Parse data from sc point.

        Parameters
        ----------
        row_idx : int
            Index value for the row of the target SC point in self._sc_build.

        Returns
        -------
        res_gids : list
            Resource GIDs associated with SC point i.
        gid_counts : list
            Number of exclusion pixels that are not excluded associated
            with each res_gid.
        buildout : float
            Total REEDS requested buildout associated with SC point i.
        capacity : float
            Total Supply Curve Point Capacity
        """
        buildout = float(self._sc_build.loc[row_idx, 'built_capacity'])
        capacity = float(self._sc_build.loc[row_idx, 'potential_capacity'])

        res_gids = self._sc_build.loc[row_idx, 'res_gids']
        gid_counts = self._sc_build.loc[row_idx, 'gid_counts']

        if isinstance(res_gids, str):
            res_gids = json.loads(res_gids)

        if isinstance(gid_counts, str):
            gid_counts = json.loads(gid_counts)

        return res_gids, gid_counts, buildout, capacity

    def _build_sc_profile(self, row_idx, profile):
        """Build a power generation profile based on SC point i.

        Parameters
        ----------
        row_idx : int
            Index value for the row of the target SC point in self._sc_build.
        profile : np.ndarray | None
            (t,) array of generation in MW, or None if this is the first
            SC point to add generation.

        Returns
        ----------
        profile : np.ndarray
            (t,) array of generation in MW where t is the timeindex length.
        res_gids : list
            List of resource GID's that were built from this SC point.
        gen_gids : list
            List of generation GID's that were built from this SC point.
        res_built : list
            List of built capacities at each resource GID from this SC point.
        """
        sc_meta, buildout = self._get_sc_meta(row_idx)

        res_gids = []
        gen_gids = []
        res_built = []

        for _, row in sc_meta.iterrows():

            if buildout <= row['gid_capacity']:
                to_build = buildout
            else:
                to_build = row['gid_capacity']

            buildout -= to_build

            res_built.append(np.round(to_build, decimals=5))

            gen_gid = int(row['gen_gid'])
            if self._forecast_map is None:
                with Outputs(self._cf_fpath, mode='r') as cf_outs:
                    cf_profile = cf_outs['cf_profile', :, gen_gid]
            else:
                gen_gid = int(self._forecast_map[gen_gid])
                with Outputs(self._forecast_fpath, mode='r') as cf_outs:
                    cf_profile = cf_outs['cf_profile', :, gen_gid]

            res_gids.append(row['res_gid'])
            gen_gids.append(gen_gid)

            if profile is None:
                profile = to_build * cf_profile
            else:
                profile += to_build * cf_profile

            if buildout <= 0:
                break

        if len(profile.shape) != 1:
            profile = profile.flatten()

        return profile, res_gids, gen_gids, res_built

    def _make_node_profile(self):
        """Make an aggregated generation profile for a single plexos node.

        Returns
        -------
        profile : np.ndarray
            (t, ) array of generation in MW.
        res_gids : list
            List of resource GID's that were built for this plexos node.
        gen_gids : list
            List of generation GID's that were built for this plexos node.
        res_built : list
            List of built capacities at each resource GID for this plexos node.
        """

        profile = None
        res_gids = []
        gen_gids = []
        res_built = []

        for i in self._sc_build.index.values:

            profile, i_res_gids, i_gen_gids, i_res_built = \
                self._build_sc_profile(i, profile)

            res_gids += i_res_gids
            gen_gids += i_gen_gids
            res_built += i_res_built

        return profile, res_gids, gen_gids, res_built

    @classmethod
    def run(cls, sc_build, cf_fpath, cf_res_gids=None, forecast_fpath=None,
            forecast_map=None):
        """Make an aggregated generation profile for a single plexos node.

        Parameters
        ----------
        sc_build : pd.DataFrame
            Supply curve buildout table. Must only have rows that are built
            in this plexos node. Must have resource_gid lookup, counts per
            resource_gid, and capacity at each SC point.
        cf_fpath : str
            File path to capacity factor file to get profiles from.
        cf_res_gids : list | np.ndarray, optional
            Resource GID's available in cf_fpath, if None pull from cf_fpath,
            by default None
        forecast_fpath : str | None, optional
            Forecasted capacity factor .h5 file path (reV results).
            If not None, the generation profiles are sourced from this file,
            by default None
        forecast_map : np.ndarray | None, optional
            (n, 1) array of forecast meta data indices mapped to the generation
            meta indices where n is the number of generation points. None if
            no forecast data being considered,
            by default None

        Returns
        -------
        profile : np.ndarray
            (t, ) array of generation in MW.
        res_gids : list
            List of resource GID's that were built for this plexos node.
        gen_gids : list
            List of generation GID's that were built for this plexos node.
        res_built : list
            List of built capacities at each resource GID for this plexos node.
        """
        n = cls(sc_build, cf_fpath, cf_res_gids=cf_res_gids,
                forecast_fpath=forecast_fpath, forecast_map=forecast_map)

        profile, res_gids, gen_gids, res_built = n._make_node_profile()

        return profile, res_gids, gen_gids, res_built


class PlexosAggregation:
    """
    Framework to aggregate reV gen profiles to PLEXOS node power profiles.
    """

    def __init__(self, plexos_nodes, rev_sc, reeds_build, cf_fpath,
                 forecast_fpath=None, build_year=2050, max_workers=None):
        """
        Parameters
        ----------
        plexos_nodes : str | pd.DataFrame
            Plexos node meta data including gid, lat/lon, voltage. Or path
            to .csv containing plexos node meta data
        rev_sc : str | pd.DataFrame
            reV supply curve results table including SC gid, lat/lon,
            res_gids, res_id_counts. Or  path to reV supply curve table.
        reeds_build : str | pd.DataFrame
            RdEDS buildout with rows for built capacity at each reV SC point.
            Or ReEDS buildout table
        cf_fpath : str
            File path to capacity factor file to get profiles from.
        forecast_fpath : str | None
            Forecasted capacity factor .h5 file path (reV results).
            If not None, the generation profiles are sourced from this file.
        build_year : int, optional
            REEDS year of interest, by default 2050
        max_workers : int | None
            Do node aggregation on max_workers.
        """
        self._plexos_nodes = self._parse_plexos_nodes(plexos_nodes)
        self._cf_fpath = cf_fpath
        self._forecast_fpath = forecast_fpath
        self.build_year = build_year
        self._cf_res_gids = None
        self._plexos_meta = None
        self._time_index = None
        self.max_workers = max_workers

        logger.info('Running PLEXOS aggregation for build year: {}'
                    .format(build_year))

        self._sc_build = self._parse_rev_reeds(rev_sc, reeds_build,
                                               build_year=build_year)
        missing = self._check_gids()
        self._handle_missing_resource_gids(missing)

        self._node_map = self._make_node_map()
        self._forecast_map = self._make_forecast_map()

    @property
    def time_index(self):
        """Get the generation profile time index.

        Returns
        -------
        time_index : pd.Datetimeindex
            Pandas datetime index sourced from the capacity factor data.
        """

        if self._time_index is None:
            with Outputs(self._cf_fpath, mode='r') as cf_outs:
                self._time_index = cf_outs.time_index

        return self._time_index

    @property
    def plexos_meta(self):
        """Get plexos node meta data for the nodes included in this problem.

        Returns
        -------
        plexos_meta : pd.DataFrame
            Plexos meta dataframe reduced to the nodes in this problem.
        """

        if self._plexos_meta is None:
            inodes = np.unique(self._node_map)

            node_builds = []
            for i in inodes:
                mask = (self._node_map == i)
                built_cap = self._sc_build[mask]['built_capacity'].values.sum()
                node_builds.append(built_cap)

            self._plexos_meta = self._plexos_nodes.iloc[inodes, :]
            self._plexos_meta['built_capacity'] = node_builds

            self._plexos_meta = DataCleaner.reduce_df(
                self._plexos_meta, DataCleaner.PLEXOS_META_COLS)

            self._plexos_meta['res_gids'] = None
            self._plexos_meta['gen_gids'] = None
            self._plexos_meta['res_built'] = None

        return self._plexos_meta

    @property
    def n_plexos_nodes(self):
        """Get the number of unique plexos nodes in this buildout.

        Returns
        -------
        n : int
            Number of unique plexos nodes in this buildout
        """
        return len(self.plexos_meta)

    @property
    def sc_res_gids(self):
        """List of unique resource GIDS in the REEDS build out.

        Returns
        -------
        sc_res_gids : np.ndarray
            Array of resource GIDs associated with this REEDS buildout.
        """

        gid_col = self._sc_build['res_gids'].values

        if isinstance(gid_col[0], str):
            gid_col = [json.loads(s) for s in gid_col]
        else:
            gid_col = list(gid_col)

        res_gids = [g for sub in gid_col for g in sub]
        sc_res_gids = np.array(sorted(list(set(res_gids))))

        return sc_res_gids

    @property
    def available_res_gids(self):
        """Resource gids available in the cf file.

        Returns
        -------
        cf_res_gids : np.ndarray
            Array of resource GIDs available in the cf file.
        """

        if self._cf_res_gids is None:
            with Outputs(self._cf_fpath, mode='r') as cf_outs:
                self._cf_res_gids = cf_outs.get_meta_arr('gid')

            if not isinstance(self._cf_res_gids, np.ndarray):
                self._cf_res_gids = np.array(list(self._cf_res_gids))

        return self._cf_res_gids

    @staticmethod
    def _parse_plexos_nodes(plexos_nodes):
        """
        Load Plexos node meta data from disc if needed, pre-filter and rename
        columns

        Parameters
        ----------
        plexos_nodes : str | pd.DataFrame
            Plexos node meta data including gid, lat/lon, voltage. Or path
            to .csv containing plexos node meta data

        Returns
        -------
        plexos_nodes : pd.DataFrame
            Plexos node meta data including gid, lat/lon, voltage
        """
        plexos_nodes = parse_table(plexos_nodes)
        plexos_nodes = DataCleaner.pre_filter_plexos_meta(plexos_nodes)
        plexos_nodes = DataCleaner.rename_cols(plexos_nodes)

        return plexos_nodes

    @staticmethod
    def _check_rev_reeds_coordinates(rev_sc, reeds_build, atol=0.5):
        """Check that the coordinates are the same in rev and reeds buildouts.

        Parameters
        ----------
        rev_sc : pd.DataFrame
            reV supply curve results table including SC gid, lat/lon,
            res_gids, res_id_counts.
        reeds_build : pd.DataFrame
            REEDS buildout with rows for built capacity at each reV SC point.
        atol : float
            Maximum difference in coord matching.

        Returns
        -------
        rev_sc : pd.DataFrame
            Same as input.
        reeds_build : pd.DataFrame
            Same as input but without lat/lon columns if matched.
        """
        join_on = 'sc_gid'
        reeds_build = reeds_build.sort_values(join_on)
        reeds_sc_gids = reeds_build[join_on].values
        rev_mask = rev_sc[join_on].isin(reeds_sc_gids)
        if not rev_mask.any():
            msg = ("There are no overlapping sc_gids between the provided reV "
                   "supply curve table the ReEDS buildout!")
            logger.error(msg)
            raise RuntimeError(msg)

        rev_sc = rev_sc.sort_values(join_on)

        rev_coord_labels = get_coord_labels(rev_sc)
        reeds_coord_labels = get_coord_labels(reeds_build)

        if rev_coord_labels is not None and reeds_coord_labels is not None:
            reeds_coords = reeds_build[reeds_coord_labels].values
            rev_coords = rev_sc.loc[rev_mask, rev_coord_labels].values

            check = np.allclose(reeds_coords, rev_coords, atol=atol, rtol=0.0)
            if not check:
                emsg = ('reV SC and REEDS Buildout coordinates do not match.')
                logger.exception(emsg)
                raise ValueError(emsg)

            reeds_build = reeds_build.drop(labels=reeds_coord_labels, axis=1)

        return rev_sc, reeds_build

    @classmethod
    def _parse_rev_reeds(cls, rev_sc, reeds_build, build_year=2050):
        """Parse and combine reV SC and REEDS buildout tables into single table

        Parameters
        ----------
        rev_sc : str | pd.DataFrame
            reV supply curve results table including SC gid, lat/lon,
            res_gids, res_id_counts. Or  path to reV supply curve table.
        reeds_build : str | pd.DataFrame
            RdEDS buildout with rows for built capacity at each reV SC point.
            Or ReEDS buildout table
        build_year : int, optional
            REEDS year of interest, by default 2050

        Returns
        -------
        table : pd.DataFrame
            rev_sc and reeds_build inner joined on supply curve gid.
        """
        rev_sc = DataCleaner.rename_cols(
            parse_table(rev_sc),
            name_map=DataCleaner.REV_NAME_MAP)
        reeds_build = DataCleaner.rename_cols(
            parse_table(reeds_build),
            name_map=DataCleaner.REEDS_NAME_MAP)

        year_mask = (reeds_build['reeds_year'] == build_year)
        if not any(year_mask):
            msg = 'Build year {} not found in reeds data!'.format(build_year)
            logger.error(msg)
            raise ValueError(msg)

        reeds_build = reeds_build[year_mask]

        join_on = 'sc_gid'
        if 'sc_gid' not in rev_sc or 'sc_gid' not in reeds_build:
            raise KeyError('GID must be in reV SC and REEDS Buildout tables!')

        rev_sc, reeds_build = cls._check_rev_reeds_coordinates(rev_sc,
                                                               reeds_build)

        check_isin = np.isin(reeds_build[join_on].values,
                             rev_sc[join_on].values)
        if not all(check_isin):
            wmsg = ('There are REEDS buildout GIDs that are not in the reV '
                    'supply curve table: {} out of {} total REEDS buildout '
                    'sites.'.format(np.sum(~check_isin), len(reeds_build)))
            warn(wmsg)
            logger.warning(wmsg)

        table = pd.merge(rev_sc, reeds_build, how='inner', left_on=join_on,
                         right_on=join_on)

        return table

    def _check_gids(self):
        """Ensure that the SC buildout GIDs are available in the cf file.

        Returns
        -------
        bad_sc_points : list
            List of missing supply curve gids
            (in reeds but not in reV resource).
        """
        bad_sc_points = []
        missing = list(set(self.sc_res_gids) - set(self.available_res_gids))
        if any(missing):
            wmsg = ('The CF file is missing {} resource gids that were built '
                    'in the REEDS-reV SC build out: {}'
                    .format(len(missing), missing))
            warn(wmsg)
            logger.warning(wmsg)

            gid_col = self._sc_build['res_gids'].values
            if isinstance(gid_col[0], str):
                gid_col = [json.loads(s) for s in gid_col]
            else:
                gid_col = list(gid_col)

            for i, sc_gids in enumerate(gid_col):
                if any([m in sc_gids for m in missing]):
                    bad_sc_points.append(self._sc_build.iloc[i]['sc_gid'])

            wmsg = ('There are {} SC points with missing gids: {}'
                    .format(len(bad_sc_points), bad_sc_points))
            warn(wmsg)
            logger.warning(wmsg)

        return bad_sc_points

    def _handle_missing_resource_gids(self, bad_sc_points):
        """Merge requested capacity in missing SC gids into nearest good pixels

        Parameters
        ----------
        bad_sc_points : list
            List of missing supply curve gids
            (in reeds but not in reV resource).
        """
        if any(bad_sc_points):
            bad_bool = self._sc_build['sc_gid'].isin(bad_sc_points)
            bad_cap_arr = self._sc_build.loc[bad_bool, 'built_capacity'].values
            good_bool = ~bad_bool
            bad_cap = bad_cap_arr.sum()
            wmsg = ('{} MW of capacity is being merged from bad SC points.'
                    .format(bad_cap))
            warn(wmsg)
            logger.warning(wmsg)

            clabels = get_coord_labels(self._sc_build)
            # pylint: disable=not-callable
            good_tree = cKDTree(self._sc_build.loc[good_bool, clabels])
            _, i = good_tree.query(self._sc_build.loc[bad_bool, clabels])

            ilen = len(self._sc_build)
            icap = self._sc_build['built_capacity'].sum()

            add_index = self._sc_build.index.values[good_bool][i]

            for i, ai in enumerate(add_index):
                self._sc_build.loc[ai, 'built_capacity'] += bad_cap_arr[i]

            bad_ind = self._sc_build.index.values[bad_bool]
            self._sc_build = self._sc_build.drop(bad_ind, axis=0)

            olen = len(self._sc_build)
            ocap = self._sc_build['built_capacity'].sum()

            wmsg = ('SC build table reduced from {} to {} rows, '
                    'capacity from {} to {} (should be the same).'
                    .format(ilen, olen, icap, ocap))
            warn(wmsg)
            logger.warning(wmsg)

            cap_error = (icap - ocap) / icap
            if cap_error > 0.001:
                msg = ('Too much capacity is being lost due to missing '
                       'resource gids! Capacity difference is {}%. '
                       'Cannot continue.'.format(cap_error * 100))
                logger.error(msg)
                raise RuntimeError(msg)

    def _init_output(self):
        """Init the output array of aggregated PLEXOS profiles.

        Returns
        -------
        output : np.ndarray
            (t, n) array of zeros where t is the timeseries length and n is
            the number of plexos nodes.
        """

        if self._forecast_fpath is None:
            with Outputs(self._cf_fpath, mode='r') as out:
                t = out.shape[0]
        else:
            with Outputs(self._forecast_fpath, mode='r') as out:
                t = out.shape[0]

        shape = (t, self.n_plexos_nodes)
        output = np.zeros(shape, dtype=np.float32)
        return output

    def _make_forecast_map(self):
        """Run ckdtree to map forecast pixels to generation pixels.

        Returns
        -------
        fmap : np.ndarray | None
            (n, 1) array of forecast meta data indices mapped to the generation
            meta indices where n is the number of generation points. None if
            no forecast filepath input.
        """

        fmap = None
        if self._forecast_fpath is not None:
            logger.info('Making KDTree from forecast data: {}'
                        .format(self._forecast_fpath))
            with Outputs(self._cf_fpath) as out:
                meta_cf = out.meta

            with Outputs(self._forecast_fpath) as out:
                meta_fo = out.meta

            clabels = get_coord_labels(meta_cf)
            tree = cKDTree(meta_fo[clabels])  # pylint: disable=not-callable
            d, fmap = tree.query(meta_cf[clabels])
            logger.info('Distance (min / mean / max) from generation pixels '
                        'to forecast pixels is: {} / {} / {}'
                        .format(d.min(), d.mean(), d.max()))

        return fmap

    def _make_node_map(self, k=1):
        """Run ckdtree to map built rev SC points to plexos nodes.

        Parameters
        ----------
        k : int
            Number of neighbors to return.

        Returns
        -------
        plx_node_index : np.ndarray
            KDTree query output, (n, k) array of plexos node indices mapped to
            the SC builds where n is the number of SC points built and k is the
            number of neighbors requested. Values are the Plexos node index.
        """

        plexos_coord_labels = get_coord_labels(self._plexos_nodes)
        sc_coord_labels = get_coord_labels(self._sc_build)
        # pylint: disable=not-callable
        tree = cKDTree(self._plexos_nodes[plexos_coord_labels])
        d, plx_node_index = tree.query(self._sc_build[sc_coord_labels], k=k)
        logger.info('Plexos Node KDTree distance min / mean / max: '
                    '{} / {} / {}'
                    .format(np.round(d.min(), decimals=3),
                            np.round(d.mean(), decimals=3),
                            np.round(d.max(), decimals=3)))

        if len(plx_node_index.shape) == 1:
            plx_node_index = plx_node_index.reshape((len(plx_node_index), 1))

        return plx_node_index

    def _ammend_plexos_meta(self, row_idx, res_gids, gen_gids, res_built):
        """Ammend the plexos meta dataframe with data about resource buildouts.

        Parameters
        ----------
        row_idx : int
            Index location to modify (iloc).
        res_gids : list
            List of resource GID's that were built for this plexos node.
        gen_gids : list
            List of generation GID's that were built for this plexos node.
        res_built : list
            List of built capacities at each resource GID for this plexos node.
        """

        index = self._plexos_meta.index.values[row_idx]

        if self._plexos_meta.loc[index, 'res_gids'] is None:
            self._plexos_meta.loc[index, 'res_gids'] = str(res_gids)
            self._plexos_meta.loc[index, 'gen_gids'] = str(gen_gids)
            self._plexos_meta.loc[index, 'res_built'] = str(res_built)

        else:
            a = json.loads(self._plexos_meta.loc[index, 'res_gids']) + res_gids
            b = json.loads(self._plexos_meta.loc[index, 'gen_gids']) + gen_gids
            c = (json.loads(self._plexos_meta.loc[index, 'res_built'])
                 + res_built)

            self._plexos_meta.loc[index, 'res_gids'] = str(a)
            self._plexos_meta.loc[index, 'gen_gids'] = str(b)
            self._plexos_meta.loc[index, 'res_built'] = str(c)

    def _make_profiles(self):
        """Make a 2D array of aggregated plexos gen profiles.

        Returns
        -------
        profiles : np.ndarray
            (t, n) array of Plexos node generation profiles where t is the
            timeseries length and n is the number of plexos nodes.
        """

        if self.max_workers != 1:
            profiles = self._make_profiles_parallel()
        else:
            profiles = self._make_profiles_serial()

        return profiles

    def _make_profiles_parallel(self):
        """Make a 2D array of aggregated plexos gen profiles in parallel.

        Returns
        -------
        profiles : np.ndarray
            (t, n) array of Plexos node generation profiles where t is the
            timeseries length and n is the number of plexos nodes.
        """
        profiles = self._init_output()
        progress = 0
        futures = {}
        loggers = [__name__, 'reVX']
        with SpawnProcessPool(max_workers=self.max_workers,
                              loggers=loggers) as exe:
            for i, inode in enumerate(np.unique(self._node_map)):
                mask = (self._node_map == inode)
                f = exe.submit(PlexosNode.run,
                               self._sc_build[mask], self._cf_fpath,
                               cf_res_gids=self.available_res_gids,
                               forecast_fpath=self._forecast_fpath,
                               forecast_map=self._forecast_map)
                futures[f] = i

            for n, f in enumerate(as_completed(futures)):
                i = futures[f]
                profile, res_gids, gen_gids, res_built = f.result()
                profiles[:, i] = profile
                self._ammend_plexos_meta(i, res_gids, gen_gids, res_built)

                current_prog = (n + 1) // (len(futures) / 100)
                if current_prog > progress:
                    progress = current_prog
                    logger.info('{} % of plexos node profiles built.'
                                .format(progress))

        return profiles

    def _make_profiles_serial(self):
        """Make a 2D array of aggregated plexos gen profiles in serial.

        Returns
        -------
        profiles : np.ndarray
            (t, n) array of Plexos node generation profiles where t is the
            timeseries length and n is the number of plexos nodes.
        """
        profiles = self._init_output()
        progress = 0
        for i, inode in enumerate(np.unique(self._node_map)):
            mask = (self._node_map == inode)
            p = PlexosNode.run(
                self._sc_build[mask], self._cf_fpath,
                cf_res_gids=self.available_res_gids,
                forecast_fpath=self._forecast_fpath,
                forecast_map=self._forecast_map)

            profile, res_gids, gen_gids, res_built = p
            profiles[:, i] = profile
            self._ammend_plexos_meta(i, res_gids, gen_gids, res_built)

            current_prog = ((i + 1)
                            // (len(np.unique(self._node_map)) / 100))
            if current_prog > progress:
                progress = current_prog
                logger.info('{} % of plexos node profiles built.'
                            .format(progress))

        return profiles

    @classmethod
    def run(cls, plexos_nodes, rev_sc, reeds_build, cf_fpath,
            forecast_fpath=None, build_year=2050, max_workers=None):
        """Run plexos aggregation.

        Parameters
        ----------
        plexos_nodes : pd.DataFrame
            Plexos node meta data including gid, lat/lon, voltage.
        rev_sc : pd.DataFrame
            reV supply curve results table including SC gid, lat/lon,
            res_gids, res_id_counts.
        reeds_build : pd.DataFrame
            REEDS buildout with rows for built capacity at each reV SC point.
        cf_fpath : str
            File path to capacity factor file to get profiles from.
        forecast_fpath : str | None
            Forecasted capacity factor .h5 file path (reV results).
            If not None, the generation profiles are sourced from this file.
        build_year : int
            REEDS year of interest.
        max_workers : int | None
            Do node aggregation on max_workers.

        Returns
        -------
        plexos_meta : pd.DataFrame
            Plexos node meta data with built capacities.
        time_index : pd.datetimeindex
            Time index for the profiles.
        profiles : np.ndarray
            Generation profile timeseries at each plexos node.
        """

        pa = cls(plexos_nodes, rev_sc, reeds_build, cf_fpath,
                 forecast_fpath=forecast_fpath, build_year=build_year,
                 max_workers=max_workers)
        profiles = pa._make_profiles()

        return pa.plexos_meta, pa.time_index, profiles


class Manager:
    """Plexos job manager."""

    def __init__(self, plexos_nodes, rev_sc, reeds_build, cf_fpath,
                 forecast_fpath=None, wait=300, db_host='gds_edit.nrel.gov',
                 db_user=None, db_pass=None, db_port=5432):
        """
        Parameters
        ----------
        plexos_nodes : str | pd.DataFrame
            Plexos node meta data (CSV file path or database.schema.name)
        rev_sc : str | pd.DataFrame
            reV supply curve results (CSV file path or database.schema.name)
        reeds_build : str | pd.DataFrame
            REEDS buildout results (CSV file path or database.schema.name)
        cf_fpath : str
            Capacity factor .h5 file path.
        forecast_fpath : str | None
            Forecasted capacity factor .h5 file path (reV results).
            If not None, the generation profiles are sourced from this file.
        wait : int
            Integer seconds to wait for DB connection to become available
            before raising exception.
        db_host : str
            Database host name.
        db_user : str
            Your database user name.
        db_pass : str
            Database password (None if your password is cached).
        db_port : int
            Database port.
        """
        self.plexos_nodes = parse_table_name(plexos_nodes, wait=wait,
                                             db_host=db_host,
                                             db_user=db_user,
                                             db_pass=db_pass,
                                             db_port=db_port)
        self.plexos_nodes = DataCleaner.pre_filter_plexos_meta(
            self.plexos_nodes)

        self.rev_sc = parse_table_name(rev_sc, wait=wait,
                                       db_host=db_host,
                                       db_user=db_user,
                                       db_pass=db_pass,
                                       db_port=db_port)
        self.reeds_build = parse_table_name(reeds_build, wait=wait,
                                            db_host=db_host,
                                            db_user=db_user,
                                            db_pass=db_pass,
                                            db_port=db_port)

        self.rev_sc = DataCleaner.rename_cols(
            self.rev_sc, name_map=DataCleaner.REV_NAME_MAP)
        self.reeds_build = DataCleaner.rename_cols(
            self.reeds_build, name_map=DataCleaner.REEDS_NAME_MAP)

        self.cf_fpath = cf_fpath
        if not os.path.exists(self.cf_fpath):
            raise FileNotFoundError('Could not find cf_fpath: {}'
                                    .format(cf_fpath))

        self.forecast_fpath = forecast_fpath
        if self.forecast_fpath is not None:
            if not os.path.exists(self.forecast_fpath):
                raise FileNotFoundError('Could not find forecast_fpath: {}'
                                        .format(forecast_fpath))

    @classmethod
    def main(cls, plexos_nodes, rev_sc, reeds_build, cf_fpath,
             forecast_fpath=None, agg_kwargs=None, wait=300,
             db_host='gds_edit.nrel.gov', db_user=None, db_pass=None,
             db_port=5432):
        """Run the Plexos pipeline for a single extent.

        Parameters
        ----------
        plexos_nodes : str | pd.DataFrame
            Plexos node meta data (CSV file path or database.schema.name)
        rev_sc : str | pd.DataFrame
            reV supply curve results (CSV file path or database.schema.name)
        reeds_build : str | pd.DataFrame
            REEDS buildout results (CSV file path or database.schema.name)
        cf_fpath : str | pd.DataFrame
            Capacity factor .h5 file path.
        forecast_fpath : str | None
            Forecasted capacity factor .h5 file path (reV results).
            If not None, the generation profiles are sourced from this file.
        agg_kwargs : dict
            Optional additional kwargs for the aggregation run.
        wait : int
            Integer seconds to wait for DB connection to become available
            before raising exception.
        db_host : str
            Database host name.
        db_user : str
            Your database user name.
        db_pass : str
            Database password (None if your password is cached).
        db_port : int
            Database port.

        Returns
        -------
        meta : pd.DataFrame
            Plexos node meta data.
        time_index : pd.Datetimeindex
            Time index.
        profiles : np.ndarray
            Plexos node generation profiles.
        """

        meta = None
        time_index = None
        profiles = None

        if agg_kwargs is None:
            agg_kwargs = {}

        logger.info('Running PLEXOS aggregation with plexos nodes input: {}'
                    .format(plexos_nodes))
        logger.info('Running PLEXOS aggregation with reV SC input: {}'
                    .format(rev_sc))
        logger.info('Running PLEXOS aggregation with REEDS input: {}'
                    .format(reeds_build))
        logger.info('Running PLEXOS aggregation with reV Gen input: {}'
                    .format(cf_fpath))
        logger.info('Running PLEXOS aggregation with forecast filepath: {}'
                    .format(forecast_fpath))

        pm = cls(plexos_nodes, rev_sc, reeds_build, cf_fpath,
                 forecast_fpath=forecast_fpath, wait=wait,
                 db_host=db_host, db_user=db_user, db_pass=db_pass,
                 db_port=db_port)

        try:
            meta, time_index, profiles = PlexosAggregation.run(
                pm.plexos_nodes, pm.rev_sc, pm.reeds_build, pm.cf_fpath,
                forecast_fpath=pm.forecast_fpath, **agg_kwargs)

        except Exception as e:
            logger.exception(e)
            raise e

        return meta, time_index, profiles

    @classmethod
    def _run_group(cls, df_group, cf_year, build_year):
        """Run a group of plexos node aggregations all belonging to the same
        final extent.

        Parameters
        ----------
        df_group : str
            DataFrame from the job_file with a common group.
        cf_year : str
            Year of the cf_fpath resource year (will be inserted if {} is in
            cf_fpath).
        build_years : list | tuple
            REEDS years to run scenarios for.

        Returns
        -------
        meta : pd.DataFrame
            Plexos node meta data.
        time_index : pd.Datetimeindex
            Time index.
        profiles : np.ndarray
            Plexos node generation profiles.
        """

        dc = None

        for i in df_group.index.values:
            plexos_nodes = df_group.loc[i, 'plexos_nodes']
            reeds_build = df_group.loc[i, 'reeds_build']
            cf_fpath = df_group.loc[i, 'cf_fpath']
            if '{}' in cf_fpath:
                cf_fpath = cf_fpath.format(cf_year)
            elif cf_year not in cf_fpath:
                warn('Specified CF year {} not present in cf file string: {}'
                     .format(cf_year, cf_fpath))

            rev_sc = df_group.loc[i, 'rev_sc']

            forecast_fpath = None
            if 'forecast_fpath' in df_group:
                forecast_fpath = df_group.loc[i, 'forecast_fpath']
                if '{}' in forecast_fpath:
                    forecast_fpath = forecast_fpath.format(cf_year)
                elif cf_year not in forecast_fpath:
                    warn('Specified CF year {} not present in ECMWF file '
                         'string: {}'.format(cf_year, forecast_fpath))

            agg_kwargs = {'build_year': build_year}
            meta, ti, profiles = cls.main(plexos_nodes, rev_sc, reeds_build,
                                          cf_fpath, agg_kwargs=agg_kwargs,
                                          forecast_fpath=forecast_fpath)

            if meta is None:
                e = ('Plexos aggregation manager failed. '
                     'PlexosAggregation.run() '
                     'failed to create a meta data object.')
                logger.error(e)
                raise RuntimeError(e)
            else:
                if dc is None:
                    dc = DataCleaner(meta, profiles)
                else:
                    dc.merge_extent(meta, profiles)

        meta, profiles = dc.merge_small()

        return meta, ti, profiles

    @classmethod
    def run(cls, job, out_dir, scenario=None, cf_year=2012,
            build_years=(2024, 2050)):
        """Run plexos node aggregation for a job file input.

        Parameters
        ----------
        job : str | pd.DataFrame
            CSV file with plexos aggregation job config. Needs the following
            columns: (scenario, group, cf_fpath, reeds_build, rev_sc,
            plexos_nodes)
        out_dir : str
            Path to an output directory.
        scenario : str | None
            Optional filter to run plexos aggregation for just one scenario in
            the job.
        cf_year : str
            Year of the cf_fpath resource year (will be inserted if {} is in
            cf_fpath).
        build_years : list | tuple | int
            REEDS years to run scenarios for.
        """

        if isinstance(job, str):
            job = pd.read_csv(job)

        job = job.where(pd.notnull(job), None)

        if isinstance(build_years, int):
            build_years = [build_years]

        if scenario is not None:
            job = job[(job['scenario'] == scenario)]

        for scenario, df_scenario in job.groupby('scenario'):
            logger.info('Running scenario "{}"'.format(scenario))
            for build_year in build_years:
                logger.info('Running build year {}'.format(build_year))
                fn_out = '{}_{}_{}.h5'.format(scenario, build_year,
                                              cf_year)
                out_fpath = os.path.join(out_dir, fn_out)

                if os.path.exists(out_fpath):
                    logger.info('Skipping exists: {}'.format(out_fpath))
                else:

                    for group, df_group in df_scenario.groupby('group'):
                        logger.info('Running group "{}"'.format(group))

                        meta, time_index, profiles = cls._run_group(df_group,
                                                                    cf_year,
                                                                    build_year)

                        logger.info('Saving result for group "{}" to file: {}'
                                    .format(group, out_fpath))

                        with Outputs(out_fpath, mode='a') as out:
                            meta = to_records_array(meta)
                            time_index = time_index.astype(str)
                            dtype = "S{}".format(len(time_index[0]))
                            time_index = np.array(time_index, dtype=dtype)
                            out._create_dset('{}/meta'.format(group),
                                             meta.shape,
                                             meta.dtype,
                                             data=meta)
                            out._create_dset('{}/time_index'.format(group),
                                             time_index.shape,
                                             time_index.dtype,
                                             data=time_index)
                            out._create_dset('{}/gen_profiles'.format(group),
                                             profiles.shape,
                                             profiles.dtype,
                                             chunks=(None, 100),
                                             data=profiles)

        logger.info('Plexos aggregation complete!')



# -*- coding: utf-8 -*-
"""reVX SQL Database handler.

Created on Thu Aug 22 08:58:37 2019

@author: gbuster
"""
import pandas as pd
import psycopg2
import getpass
import time


class Database:
    """Framework to interact with the reV/reVX database."""

    def __init__(self, db_name, db_host='gds_edit.nrel.gov', db_user=None,
                 db_pass=None, db_port=5432):
        """
        Parameters
        ----------
        db_name : str
            Database name.
        db_host : str
            Database host name.
        db_user : str
            Your database user name.
        db_pass : str
            Database password (None if your password is cached).
        db_port : int
            Database port.
        """

        self.db_host = db_host or 'localhost'
        self.db_name = db_name or db_user or getpass.getuser()
        self.db_user = db_user or getpass.getuser()
        self.db_pass = db_pass
        self.db_port = db_port or 5432

        self._con = psycopg2.connect(host=self.db_host,
                                     port=self.db_port,
                                     user=self.db_user,
                                     password=self.db_pass,
                                     database=self.db_name)

    @property
    def con(self):
        """Return the database connection object."""
        return self._con

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """close DB connection on exit."""
        self._con.close()

    @classmethod
    def get_table(cls, table, schema, db_name, wait=300,
                  db_host='gds_edit.nrel.gov', db_user=None, db_pass=None,
                  db_port=5432):
        """Get a table using a database query.

        Parameters
        ----------
        table : str
            Table name to retrieve from schema.
        schema : str
            Schema name being accessed in the database.
        db_name : str
            Database name.
        wait : int
            Integer seconds to wait for DB connection to become available
            before raising exception.
        db_host : str
            Database host name.
        db_user : str
            Your database user name.
        db_pass : str
            Database password (None if your password is cached).
        db_port : int
            Database port.

        Returns
        -------
        df : pd.DataFrame
            Dataframe representation of schema.table.
        """
        sql = 'SELECT * FROM "{}"."{}"'.format(schema, table)
        kwargs = {"db_host": db_host, "db_user": db_user,
                  "db_pass": db_pass, "db_port": db_port}

        if wait > 0:
            waited = 0
            while True:
                try:
                    with cls(db_name, **kwargs) as db:
                        df = pd.read_sql(sql, db.con)
                except psycopg2.OperationalError as e:
                    if waited > wait:
                        raise e
                    else:
                        waited += 10
                        time.sleep(10)
                else:
                    break
        else:
            with cls(db_name, **kwargs) as db:
                df = pd.read_sql(sql, db.con)

        return df



# -*- coding: utf-8 -*-
"""
Convert setback geotiff to .h5 exclusion layers
"""
import logging
import numpy as np
import os
from warnings import warn

from reVX.handlers.geotiff import Geotiff
from reVX.handlers.outputs import Outputs
from reVX.utilities.exclusions_converter import ExclusionsConverter

logger = logging.getLogger(__name__)


class SetbacksConverter(ExclusionsConverter):
    """
    Convert setbacks goetiff(s) to excl .h5 layers
    """
    @classmethod
    def _parse_setbacks(cls, setbacks, chunks=(128, 128)):
        """
        Load setbacks, combine multiple setbacks by state if needed

        Parameters
        ----------
        setbacks : list
            List of paths to setback geotiffs to load and combine
        chunks : tuple, optional
            Chunk size of exclusions in Geotiff, by default (128, 128)

        Returns
        -------
        values : ndarray
            Setbacks exclusion array
        """
        values = None
        for geotiff in setbacks:
            v = cls._parse_tiff(geotiff, chunks=chunks, check_tiff=False)[1]
            if values is None:
                values = v
            else:
                values = np.maximum(values, v)

        return values

    def setbacks_to_layer(self, layer, setbacks, check_tiff=True,
                          transform_atol=0.01, coord_atol=0.001,
                          description=None, scale_factor=None, dtype='uint8'):
        """
        Transfer geotiff exclusions to h5 confirming they match existing layers

        Parameters
        ----------
        layer : str
            Layer to extract
        geotiff : str
            Path to geotiff file
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001
        description : str, optional
            Description of exclusion layer, by default None
        scale_factor : int | float, optional
            Scale factor to use to scale geotiff data when added to the .h5
            file, by default None
        dtype : str, optional
            Dtype to save geotiff data as in the .h5 file. Only used when
            'scale_factor' is not None, by default 'uint8'
        """
        if os.path.isdir(setbacks):
            logger.debug('\t- Combining setbacks in {}'.format(setbacks))
            setbacks = [os.path.join(setbacks, file)
                        for file in os.listdir(setbacks)
                        if file.endswith('.tif')]
        else:
            logger.debug('\t- Loading setbacks from {}'.format(setbacks))
            setbacks = [setbacks]

        if not os.path.exists(self._excl_h5):
            self._init_h5(self._excl_h5, setbacks[0], chunks=self._chunks)

        msg = ("{} is already present in {}".format(layer, self._excl_h5))
        if layer in self.layers:
            if self._replace:
                msg += " and will be replaced"
                logger.warning(msg)
                warn(msg)
            else:
                msg += ", to 'replace' set to True"
                logger.error(msg)
                raise KeyError(msg)

        if check_tiff:
            self._check_geotiff(self._excl_h5, setbacks[0],
                                chunks=self._chunks,
                                transform_atol=transform_atol,
                                coord_atol=coord_atol)

        with Geotiff(setbacks[0], chunks=self._chunks) as tif:
            profile = tif.profile

        setbacks = self._parse_setbacks(setbacks, chunks=self._chunks)
        if scale_factor is not None:
            setbacks = Outputs._check_data_dtype(setbacks, dtype,
                                                 scale_factor=scale_factor)

        self._write_layer(self._excl_h5, layer, profile, setbacks,
                          chunks=self._chunks, description=description,
                          scale_factor=scale_factor)

    @classmethod
    def layers_to_h5(cls, excl_h5, layers, chunks=(128, 128),
                     replace=True, check_tiff=True,
                     transform_atol=0.01, coord_atol=0.001,
                     descriptions=None, scale_factors=None):
        """
        Create exclusions .h5 file, or load layers into existing exclusion .h5
        file from provided setbacks

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing or to contain exclusion layers
        layers : dict
            Dictionary mapping goetiffs to the layers to load
        chunks : tuple, optional
            Chunk size of exclusions in Geotiff, by default (128, 128)
        replace : bool, optional
            Flag to replace existing layers if needed, by default True
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001
        description : dict, optional
            Description of exclusion layers, by default None
        scale_factor : dict, optional
            Scale factors and dtypes to use when scaling given layers,
            by default None
        """
        if isinstance(layers, list):
            layers = {os.path.basename(lyr).split('.')[0]: lyr
                      for lyr in layers}

        if scale_factors is None:
            scale_factors = {}

        if descriptions is None:
            descriptions = {}

        excls = cls(excl_h5, chunks=chunks, replace=replace)
        logger.info('Creating {}'.format(excl_h5))
        for layer, setbacks in layers.items():
            logger.info('- Transfering {}'.format(layer))
            scale = scale_factors.get(layer, None)
            if scale is not None:
                scale_factor = scale['scale_factor']
                dtype = scale['dtype']
            else:
                scale_factor = None
                dtype = None

            description = descriptions.get(layer, None)

            excls.setbacks_to_layer(layer, setbacks, check_tiff=check_tiff,
                                    transform_atol=transform_atol,
                                    coord_atol=coord_atol,
                                    description=description,
                                    scale_factor=scale_factor,
                                    dtype=dtype)



# -*- coding: utf-8 -*-
"""
Utility to extract a years data from multi-year output .h5 files
"""
import h5py
import logging
import os

from rex.utilities.utilities import parse_year

logger = logging.getLogger(__name__)


def output_extractor(my_fpath, out_fpath, year=None):
    """
    Extract all datasets for a give year from multi-year output file

    Parameters
    ----------
    my_fpath : str
        Path to multi-year output .h5 file
    out_fpath : str
        Path to output .h5 file
    year : int | str, optional
        Year to extract, if None parse from out_fpath, by default None
    """
    if year is None:
        year = parse_year(os.path.basename(out_fpath))

    if not isinstance(year, str):
        year = str(year)

    logger.info('Transfering all datasets for {} from {} to {}'
                .format(year, my_fpath, out_fpath))
    with h5py.File(out_fpath, 'w-') as f_out:
        with h5py.File(my_fpath, 'r') as f_src:
            logger.debug('Transfering global attrs')
            for k, v in f_src.attrs.items():
                f_out.attrs[k] = v

            logger.debug('Transfering meta')
            src = f_src['meta']
            out = f_out.create_dataset('meta', shape=src.shape,
                                       dtype=src.dtype,
                                       chunks=src.chunks,
                                       data=src[...])
            for k, v in src.attrs.items():
                out.attrs[k] = v

            for dset in f_src:
                if dset.endswith(year):
                    out_name = dset.rstrip('-{}'.format(year))
                    logger.debug('Transfering {}'.format(out_name))
                    src = f_src[dset]
                    out = f_out.create_dataset(out_name,
                                               shape=src.shape,
                                               dtype=src.dtype,
                                               chunks=src.chunks,
                                               data=src[...])
                    for k, v in src.attrs.items():
                        out.attrs[k] = v

    logger.info('{} created successfully'.format(out_fpath))



# -*- coding: utf-8 -*-
"""
reVX WindSetbacks Configuration
"""

from reV.config.base_analysis_config import AnalysisConfig


class WindSetbacksConfig(AnalysisConfig):
    """Config framework for wind setbacks"""

    NAME = 'WindSetbacks'
    REQUIREMENTS = ('excl_h5', 'hub_height', 'rotor_diameter', 'features_path',
                    'feature_type')
    FEATURE_TYPES = ['structure', 'road', 'rail', 'transmission']

    @property
    def feature_type(self):
        """
        Get the setback feature type (required).
        must be one of ['structure', 'road', 'rail', 'transmission']
        """
        feature_type = self['feature_type']
        msg = ("feature_type must be one of: {}".format(self.FEATURE_TYPES))
        assert feature_type in self.FEATURE_TYPES, msg

        return feature_type

    @property
    def excl_h5(self):
        """Get the exclusions .h5 file path (required)."""
        return self['excl_h5']

    @property
    def hub_height(self):
        """Get turbine hub height (required)."""
        return self['hub_height']

    @property
    def rotor_diameter(self):
        """Get turbine rotor diameter (required)."""
        return self['rotor_diameter']

    @property
    def features_path(self):
        """Get path to features file or directory (required)."""
        return self['features_path']

    @property
    def regs_fpath(self):
        """Get wind regulations .csv path"""
        return self.get('regs_fpath', None)

    @property
    def multiplier(self):
        """Get generic max-tip height multiplier"""
        return self.get('multiplier', None)

    @property
    def max_workers(self):
        """Get maximum number of workers to use to compute setbacks"""
        return self.get('max_workers', None)

    @property
    def replace(self):
        """Get replace flag"""
        return self.get('replace', False)

    @property
    def hsds(self):
        """Get hsds flag"""
        return self.get('hsds', False)



# -*- coding: utf-8 -*-
"""
RED-E reV based tech potential tool
"""
import logging
from reV.supply_curve.exclusions import ExclusionMaskFromDict

logger = logging.getLogger(__name__)


class TechPotential(ExclusionMaskFromDict):
    """
    RED-E Tech Potential tool
    """
    def __init__(self, h5_path, base_layer, excl_dict, power_density=1,
                 hsds=False, min_area=None, kernel='queen',
                 check_layers=False):
        """
        Parameters
        ----------
        h5_path : str
            Path to .h5 file containing CF means and exclusion layers
        base_layer : str
            Name of dataset in .h5 file containing base layer
        excl_dict : dcit
            Dictionary of exclusion LayerMask arugments {layer: {kwarg: value}}
        power_density : float
            Multiplier to convert CF means to generation means
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        min_area : float | NoneType
            Minimum required contiguous area in sq-km
        kernel : str
            Contiguous filter method to use on final exclusion
        check_layers : bool
            Run a pre-flight check on each layer to ensure they contain
            un-excluded values
        """
        layers_dict = excl_dict.copy()
        layers_dict[base_layer] = {"use_as_weights": True}
        super().__init__(h5_path, layers_dict=layers_dict, hsds=hsds,
                         min_area=min_area, kernel=kernel,
                         check_layers=check_layers)
        self._pd = power_density

    @property
    def profile(self):
        """
        GeoTiff profile for exclusions

        Returns
        -------
        profile : dict
            Generic GeoTiff profile for exclusions in .h5 file
        """
        return self.excl_h5.profile

    @property
    def generation(self):
        """
        Tech-potential as generation

        Returns
        -------
        gen : ndarray
        """
        gen = self[...]

        return gen * self._pd

    @staticmethod
    def _parse_sub_slice(sub_slice):
        """
        Convert sub_slice tuples into a tuple of slice objects

        Parameters
        ----------
        sub_slice : tuple | None
            Tuple of sub-window coordinates in the format (start, end, step),
            if None return (slice(None), )

        Returns
        -------
        out_slice : tuple
            Tuple of equivalent slice objects
        """
        if sub_slice is not None:
            if not isinstance(sub_slice[0], (list, tuple)):
                sub_slice = (sub_slice, )

            out_slice = ()
            for s in sub_slice:
                out_slice += (slice(*s), )
        else:
            out_slice = (slice(None), )

        return out_slice

    @classmethod
    def run(cls, h5_path, base_layer, excl_dict, sub_slice=None,
            power_density=1, hsds=False, min_area=None, kernel='queen',
            check_layers=False, generation=False):
        """
        compute tech-potential

        Parameters
        ----------
        h5_path : str
            Path to .h5 file containing CF means and exclusion layers
        base_layer : str
            Name of dataset in .h5 file containing base layer
        excl_dict : dcit
            Dictionary of exclusion LayerMask arugments {layer: {kwarg: value}}
        sub_slice : tuple, optional
            Coordinates of sub-slice to extract, if None extract entire mask,
            by default None
        power_density : float, optional
            Multiplier to convert CF means to generation means, by default 1
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        min_area : float | NoneType, optional
            Minimum required contiguous area in sq-km, by default None
        kernel : str, optional
            Contiguous filter method to use on final exclusion,
            by default 'queen'
        check_layers : bool, optional
            Run a pre-flight check on each layer to ensure they contain
            un-excluded values, by default False
        generation : bool, optional
            Flag to return generation instead of base layer:
            gen = base layer * power density
            by default False

        Returns
        -------
        out : ndarray
            Base layer with exclusions masked out, if generation is true
            multiply base layer mask by power density
        """
        with cls(h5_path, base_layer, excl_dict, power_density=power_density,
                 hsds=hsds, min_area=min_area, kernel=kernel,
                 check_layers=check_layers) as f:
            sub_slice = cls._parse_sub_slice(sub_slice)
            out = f[sub_slice]

            if generation:
                out *= f._pd

        return out



# -*- coding: utf-8 -*-
"""
Compute wind setbacks exclusions
"""
from abc import ABC
from concurrent.futures import as_completed
import fiona
import geopandas as gpd
import logging
import os
from rasterio import features
import re
from shapely.geometry import shape
from warnings import warn

from rex.utilities import parse_table, SpawnProcessPool, log_mem
from reV.handlers.exclusions import ExclusionLayers
from reVX.utilities.exclusions_converter import ExclusionsConverter

logger = logging.getLogger(__name__)


class BaseWindSetbacks(ABC):
    """
    Create exclusions layers for wind setbacks
    """
    MULTIPLIERS = {'high': 3, 'moderate': 1.1}

    def __init__(self, excl_h5, hub_height, rotor_diameter, regs_fpath=None,
                 multiplier=None, hsds=False, chunks=(128, 128)):
        """
        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers
        hub_height : float | int
            Turbine hub height (m), used along with rotor diameter to compute
            blade tip height which is used to determine setback distance
        rotor_diameter : float | int
            Turbine rotor diameter (m), used along with hub height to compute
            blade tip height which is used to determine setback distance
        regs_fpath : str | None, optional
            Path to wind regulations .csv file, if None create generic
            setbacks using max-tip height * "multiplier", by default None
        multiplier : int | float | str | None, optional
            setback multiplier to use if wind regulations are not supplied,
            if str, must a key in {'high': 3, 'moderate': 1.1}, if supplied
            along with regs_fpath, will be ignored, multiplied with
            max-tip height, by default None
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        chunks : tuple, optional
            Chunk size to use for setback layers, if None use default chunk
            size in excl_h5, by default (128, 128)
        """
        self._excl_h5 = excl_h5
        self._hub_height = hub_height
        self._rotor_diameter = rotor_diameter
        self._hsds = hsds
        self._shape, self._chunks, self._profile = \
            self._parse_excl_properties(excl_h5, chunks, hsds=hsds)

        self._regs, self._multi = self._preflight_check(regs_fpath, multiplier)

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self._excl_h5)
        return msg

    @property
    def hub_height(self):
        """
        Turbine hub-height in meters

        Returns
        -------
        float
        """
        return self._hub_height

    @property
    def rotor_diameter(self):
        """
        Turbine rotor diameter in meters

        Returns
        -------
        float
        """
        return self._rotor_diameter

    @property
    def tip_height(self):
        """
        Turbine blade tip height in meters

        Returns
        -------
        float
        """
        return self._hub_height + self._rotor_diameter / 2

    @property
    def generic_setback(self):
        """
        Default setback of turbine tip height * multiplier, used for global
        setbacks

        Returns
        -------
        float
        """
        if self._multi is None:
            setback = None
        else:
            setback = self.tip_height * self._multi

        return setback

    @property
    def wind_regs(self):
        """
        Wind Regulations table

        Returns
        -------
        geopands.GeoDataFrame | None
        """
        return self._regs

    @property
    def multiplier(self):
        """
        Generic setback multiplier

        Returns
        -------
        int | float
        """
        return self._multi

    @property
    def arr_shape(self):
        """
        Rasterize array shape

        Returns
        -------
        tuple
        """
        return self._shape

    @property
    def profile(self):
        """
        Geotiff profile

        Returns
        -------
        dict
        """
        return self._profile

    @property
    def crs(self):
        """
        Coordinate reference system

        Returns
        -------
        str
        """
        return self.profile['crs']

    @staticmethod
    def _parse_excl_properties(excl_h5, chunks, hsds=False):
        """
        Parse exclusions shape, chunk size, and profile from excl_h5 file

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers
        chunks : tuple | None
            Chunk size of exclusions datasets
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False

        Returns
        -------
        shape : tuple
            Shape of exclusions datasets
        chunks : tuple | None
            Chunk size of exclusions datasets
        profile : str
            GeoTiff profile for exclusions datasets
        """
        with ExclusionLayers(excl_h5, hsds=hsds) as exc:
            shape = exc.shape
            profile = exc.profile
            if chunks is None:
                chunks = exc.chunks

        if len(chunks) < 3:
            chunks = (1, ) + chunks

        if len(shape) < 3:
            shape = (1, ) + shape

        logger.debug('Exclusions properties:\n'
                     'shape : {}\n'
                     'chunks : {}\n'
                     'profile : {}\n'
                     .format(shape, chunks, profile))

        return shape, chunks, profile

    @staticmethod
    def _get_setback(cnty_regs, hub_height, rotor_diameter):
        """
        Compute the setback distance in meters from the county regulations,
        turbine tip height or rotor diameter

        Parameters
        ----------
        cnty_regs : pandas.Series
            Pandas Series with wind regulations for a single county / feature
            type
        hub_height : int | float
            Turbine hub-height, used to compute setbacks from hub-height
            multiplier regulations and to compute tip-height for tip-height
            multiplier regulations
        rotor_diameter : int | float
            Turbine rotor diameter, used to compute setbacks from
            rotor-diameter multiplier regulations

        Returns
        -------
        setback : float | None
            setback distance in meters, None if the setback "Value Type"
            was not recognized
        """
        tip_height = hub_height + rotor_diameter / 2

        setback_type = cnty_regs['Value Type']
        setback = cnty_regs['Value']
        if setback_type == 'Max-tip Height Multiplier':
            setback *= tip_height
        elif setback_type == 'Rotor-Diameter Multiplier':
            setback *= rotor_diameter
        elif setback_type == 'Hub-height Multiplier':
            setback *= hub_height
        elif setback_type != 'Meters':
            msg = ('Cannot create setback for {}, expecting '
                   '"Max-tip Height Multiplier", '
                   '"Rotor-Diameter Multiplier", '
                   '"Hub-height Multiplier", or '
                   '"Meters", but got {}'
                   .format(cnty_regs['County'], setback_type))
            logger.warning(msg)
            warn(msg)
            setback = None

        return setback

    @staticmethod
    def _parse_features(features_fpath, crs):
        """
        Abstract method to parse features

        Parameters
        ----------
        features_fpath : str
            Path to file containing features to setback from
        crs : str
            Coordinate reference system to convert structures geometries into

        Returns
        -------
        features : geopandas.GeoDataFrame
            Geometries of features to setback from in exclusion coordinate
            system
        """
        features = gpd.read_file(features_fpath)

        return features.to_crs(crs=crs)

    @staticmethod
    def _compute_local_setbacks(features, cnty, setback):
        """
        Compute local features setbacks

        Parameters
        ----------
        features : geopandas.GeoDataFrame
            Features to setback from
        cnty : geopandas.GeoDataFrame
            Wind regulations for a single county
        setback : int
            Setback distance in meters

        Returns
        -------
        setbacks : list
            List of setback geometries
        """
        logger.debug('- Computing setbacks for county FIPS {}'
                     .format(cnty.iloc[0]['FIPS']))
        log_mem(logger)
        mask = features.centroid.within(cnty['geometry'].values[0])
        tmp = features.loc[mask]
        tmp.loc[:, 'geometry'] = tmp.buffer(setback)

        setbacks = [(geom, 1) for geom in tmp['geometry']]

        return setbacks

    def _parse_county_regs(self, regs):
        """
        Parse wind regulations, combine with county geometries from
        exclusions .h5 file. The county geometries are intersected with
        features to compute county specific setbacks.

        Parameters
        ----------
        regs : pandas.DataFrame
            Wind regulations table

        Returns
        -------
        regs: geopandas.GeoDataFrame
            GeoDataFrame with county level wind setback regulations merged
            with county geometries, use for intersecting with setback features
        """
        if 'FIPS' not in regs:
            msg = ('Wind regulations does not have county FIPS! Please add a '
                   '"FIPS" columns with the unique county FIPS values.')
            logger.error(msg)
            raise RuntimeError(msg)

        if 'geometry' not in regs:
            regs['geometry'] = None

        regs = gpd.GeoDataFrame(regs, crs=self.crs, geometry='geometry')
        regs = regs.set_index('FIPS')

        logger.info('Merging county geometries w/ local wind '
                    'regulations')
        with ExclusionLayers(self._excl_h5) as exc:
            fips = exc['cnty_fips']
            profile = exc.get_layer_profile('cnty_fips')

        tr = profile['transform']

        for p, v in features.shapes(fips, transform=tr):
            v = int(v)
            if v in regs.index:
                regs.at[v, 'geometry'] = shape(p)

        return regs.reset_index()

    def _parse_regs(self, regs_fpath):
        """
        Parse wind regulations

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file

        Returns
        -------
        regs: geopandas.GeoDataFrame
            GeoDataFrame with county level wind setback regulations merged
            with county geometries, use for intersecting with setback features
        """
        try:
            regs = parse_table(regs_fpath)
            regs = self._parse_county_regs(regs)
            log_mem(logger)

            out_path = regs_fpath.split('.')[0] + '.gpkg'
            logger.debug('Saving wind regulations with county geometries as: '
                         '{}'.format(out_path))
            regs.to_file(out_path, driver='GPKG')
        except ValueError:
            regs = gpd.read_file(regs_fpath)

        fips_check = regs['geometry'].isnull()
        if fips_check.any():
            msg = ('The following county FIPS were requested in the '
                   'wind regulations but were not availble in the '
                   'Exclusions "cnty_fips" layer:\n{}'
                   .format(regs.loc[fips_check, 'FIPS']))
            logger.error(msg)
            raise RuntimeError(msg)

        return regs.to_crs(crs=self.crs)

    def _preflight_check(self, regs_fpath, multiplier):
        """
        Run preflight checks on WindSetBack inputs:
        1) Ensure either a wind regulations .csv is provided, or
           a setback multiplier
        2) Ensure wind regulations has county FIPS, map regulations to county
           geometries from exclusions .h5 file
        3) Ensure multiplier is a valid entry, either a float or one of
           {'high': 3, 'moderate': 1.1}

        Parameters
        ----------
        regs_fpath : str | None
            Path to wind regulations .csv file, if None create global
            setbacks
        multiplier : int | float | str | None
            setback multiplier to use if wind regulations are not supplied,
            if str, must one of {'high': 3, 'moderate': 1.1}

        Returns
        -------
        regs: geopandas.GeoDataFrame | None
            GeoDataFrame with county level wind setback regulations merged
            with county geometries, use for intersecting with setback features
        Multiplier : float | None
            Generic setbacks multiplier
        """
        if regs_fpath:
            if multiplier:
                msg = ('A wind regulation .csv file was also provided and '
                       'will be used to determine setback multipliers!')
                logger.warning(msg)
                warn(msg)

            multiplier = None
            regs = self._parse_regs(regs_fpath)
            logger.debug('Computing setbacks using regulations provided in: {}'
                         .format(regs_fpath))
        elif multiplier:
            regs = None
            if isinstance(multiplier, str):
                multiplier = self.MULTIPLIERS[multiplier]

            logger.debug('Computing setbacks using generic Max-tip Height '
                         'Multiplier of {}'.format(multiplier))
        else:
            msg = ('Computing setbacks requires either a wind regulations '
                   '.csv file or a generic multiplier!')
            logger.error(msg)
            raise RuntimeError(msg)

        return regs, multiplier

    def _check_regs(self, features_fpath):  # pylint: disable=unused-argument
        """
        Reduce regs to state corresponding to features_fpath if needed

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from

        Returns
        -------
        regs : geopands.GeoDataFrame | None
            Wind Regulations
        """
        logger.debug('Computing setbacks for wind regulations in {} counties'
                     .format(len(self.wind_regs)))

        return self.wind_regs

    def _rasterize_setbacks(self, shapes):
        """
        Convert setbacks geometries into exclusions array

        Parameters
        ----------
        setbacks : list
            List of (geometry, 1) pairs to rasterize. Each geometry is a
            feature buffered by the desired setback distance in meters

        Returns
        -------
        arr : ndarray
            Rasterized array of setbacks
        """
        logger.debug('Generating setbacks exclusion array of shape {}'
                     .format(self.arr_shape))
        arr = features.rasterize(shapes=shapes,
                                 out_shape=self.arr_shape[1:],
                                 fill=0,
                                 transform=self.profile['transform'],
                                 dtype='uint8')

        return arr

    def _write_setbacks(self, geotiff, setbacks, replace=False):
        """
        Write setbacks to geotiff, replace if requested

        Parameters
        ----------
        geotiff : str
            Path to geotiff file to save setbacks too
        setbacks : ndarray
            Rasterized array of setbacks
        replace : bool, optional
            Flag to replace local layer data with arr if layer already
            exists in the exlcusion .h5 file, by default False
        """
        if os.path.exists(geotiff):
            if not replace:
                msg = ('{} already exists. To replace it set "replace=True"'
                       .format(geotiff))
                logger.error(msg)
                raise IOError(msg)
            else:
                msg = ('{} already exists and will be replaced!'
                       .format(geotiff))
                logger.warning(msg)
                warn(msg)

        ExclusionsConverter._write_geotiff(geotiff, self.profile, setbacks)

    def compute_local_setbacks(self, features_fpath, max_workers=None):
        """
        Compute local setbacks for all counties either in serial or parallel.

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        wind_regs = self._check_regs(features_fpath)
        setbacks = []
        features = self._parse_features(features_fpath, self.crs)
        if max_workers is None:
            max_workers = os.cpu_count()

        log_mem(logger)
        if max_workers > 1:
            logger.info('Computing local setbacks in parallel using {} '
                        'workers'.format(max_workers))
            loggers = [__name__, 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for i in range(len(wind_regs)):
                    cnty = wind_regs.iloc[[i]].copy()
                    setback = self._get_setback(cnty.iloc[0], self.hub_height,
                                                self.rotor_diameter)
                    if setback is not None:
                        idx = features.sindex.intersection(cnty.total_bounds)
                        cnty_feats = features.iloc[list(idx)].copy()
                        future = exe.submit(self._compute_local_setbacks,
                                            cnty_feats, cnty, setback)
                        futures.append(future)

                for i, future in enumerate(as_completed(futures)):
                    setbacks.extend(future.result())
                    logger.debug('Computed setbacks for {} of {} counties'
                                 .format((i + 1), len(wind_regs)))
        else:
            logger.info('Computing local setbacks in serial')
            for i in range(len(wind_regs)):
                cnty = wind_regs.iloc[[i]]
                setback = self._get_setback(cnty.iloc[0], self.hub_height,
                                            self.rotor_diameter)
                if setback is not None:
                    idx = features.sindex.intersection(cnty.total_bounds)
                    cnty_feats = features.iloc[list(idx)]
                    setbacks.extend(self._compute_local_setbacks(cnty_feats,
                                                                 cnty,
                                                                 setback))
                    logger.debug('Computed setbacks for {} of {} counties'
                                 .format((i + 1), len(wind_regs)))

        return self._rasterize_setbacks(setbacks)

    def compute_generic_setbacks(self, features_fpath):
        """
        Compute generic setbacks.

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        logger.info('Computing generic setbacks')
        features = self._parse_features(features_fpath, self.crs)
        features.loc[:, 'geometry'] = features.buffer(self.generic_setback)
        setbacks = [(geom, 1) for geom in features['geometry']]

        return self._rasterize_setbacks(setbacks)

    def compute_setbacks(self, features_fpath, max_workers=None,
                         geotiff=None, replace=False):
        """
        Compute setbacks for all states either in serial or parallel.
        Existing setbacks are computed if a wind regulations file was supplied
        during class initialization, otherwise generic setbacks are computed

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None
        geotiff : str, optional
            Path to save geotiff containing rasterized setbacks,
            by default None
        replace : bool, optional
            Flag to replace geotiff if it already exists, by default False

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        if self._regs is not None:
            setbacks = self.compute_local_setbacks(features_fpath,
                                                   max_workers=max_workers)
        else:
            setbacks = self.compute_generic_setbacks(features_fpath)

        if geotiff is not None:
            logger.debug('Writing setbacks to {}'.format(geotiff))
            self._write_setbacks(geotiff, setbacks, replace=replace)

        return setbacks


class StructureWindSetbacks(BaseWindSetbacks):
    """
    Structure Wind setbacks
    """
    @staticmethod
    def _split_state_name(state_name):
        """
        Split state name at capitals to map .geojson files to regulations
        state names

        Parameters
        ----------
        state_name : str
            State name from geojson files paths with out spaces

        Returns
        -------
        str
            State names with spaces added between Capitals (names) to match
            wind regulations state names
        """
        state_name = ' '.join(a for a
                              in re.split(r'([A-Z][a-z]*)', state_name)
                              if a)

        return state_name

    @staticmethod
    def _get_feature_paths(structures_path):
        """
        Find all structures .geojson files in structures dir

        Parameters
        ----------
        structure_path : str
            Path to structures geojson for a single state, or directory
            containing geojsons for all states. Used to identify structures to
            build setbacks from. Files should be by state

        Returns
        -------
        file_paths : list
            List of file paths to all structures .geojson files in
            structures_dir
        """
        if structures_path.endswith('.geojson'):
            file_paths = [structures_path]
        else:
            file_paths = []
            for file in sorted(os.listdir(structures_path)):
                if file.endswith('.geojson'):
                    file_paths.append(os.path.join(structures_path, file))

        return file_paths

    def _parse_regs(self, regs_fpath):
        """
        Parse wind regulations, reduce table to just structures

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file

        Returns
        -------
        regs : pandas.DataFrame
            Wind regulations table
        """
        regs = super()._parse_regs(regs_fpath)

        mask = ((regs['Feature Type'] == 'Structures')
                & (regs['Comment'] != 'Occupied Community Buildings'))
        regs = regs.loc[mask]

        return regs

    def _check_regs(self, features_fpath):
        """
        Reduce regs to state corresponding to features_fpath if needed

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from

        Returns
        -------
        wind_regs : geopands.GeoDataFrame | None
            Wind Regulations
        """
        state_name = os.path.basename(features_fpath).split('.')[0]
        state = self._split_state_name(state_name)
        wind_regs = self.wind_regs
        mask = wind_regs["State"] == state

        if not mask.any():
            msg = ("There are no local regulations in {}!".format(state))
            logger.error(msg)
            raise RuntimeError(msg)

        wind_regs = wind_regs.loc[mask].reset_index(drop=True)
        logger.debug('Computing setbacks for wind regulations in {} counties'
                     .format(len(wind_regs)))

        return wind_regs

    @classmethod
    def run(cls, excl_h5, structures_path, out_dir, hub_height,
            rotor_diameter, regs_fpath=None, multiplier=None,
            chunks=(128, 128), max_workers=None, replace=False, hsds=False):
        """
        Compute state's structural setbacks and write them to a geotiff.
        If a wind regulations file is given compute local setbacks, otherwise
        compute generic setbacks using the given multiplier and the turbine
        tip-height.

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers
        structure_path : str
            Path to structures geojson for a single state, or directory
            containing geojsons for all states.
        out_dir : str
            Directory to save setbacks geotiff(s) into
        hub_height : float | int
            Turbine hub height (m), used along with rotor diameter to compute
            blade tip height which is used to determine setback distance
        rotor_diameter : float | int
            Turbine rotor diameter (m), used along with hub height to compute
            blade tip height which is used to determine setback distance
        regs_fpath : str | None, optional
            Path to wind regulations .csv file, if None create generic
            setbacks using max-tip height * "multiplier", by default None
        multiplier : int | float | str | None, optional
            setback multiplier to use if wind regulations are not supplied,
            if str, must a key in {'high': 3, 'moderate': 1.1}, if supplied
            along with regs_fpath, will be ignored, multiplied with
            max-tip height, by default None
        chunks : tuple, optional
            Chunk size to use for setback layers, if None use default chunk
            size in excl_h5, by default (128, 128)
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None
        replace : bool, optional
            Flag to replace geotiff if it already exists, by default False
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        """
        setbacks = cls(excl_h5, hub_height, rotor_diameter,
                       regs_fpath=regs_fpath, multiplier=multiplier,
                       hsds=hsds, chunks=chunks)

        structures_path = setbacks._get_feature_paths(structures_path)

        for fpath in structures_path:
            geotiff = os.path.basename(fpath).replace('.geojson', '.tif')
            geotiff = os.path.join(out_dir, geotiff)
            if os.path.exists(geotiff) and not replace:
                msg = ('{} already exists, setbacks will not be re-computed '
                       'unless replace=True'.format(geotiff))
                logger.error(msg)
            else:
                logger.info("Computing setbacks from structures in {} and "
                            "saving to {}".format(fpath, geotiff))
                setbacks.compute_setbacks(fpath, geotiff=geotiff,
                                          max_workers=max_workers,
                                          replace=replace)


class RoadWindSetbacks(BaseWindSetbacks):
    """
    Road Wind setbacks
    """
    @staticmethod
    def _parse_features(roads_fpath, crs):
        """
        Load roads from gdb file, convert to exclusions coordinate system

        Parameters
        ----------
        roads_fpath : str
            Path to here streets gdb file for given state
        crs : str
            Coordinate reference system to convert structures geometries into

        Returns
        -------
        roads : geopandas.GeoDataFrame.sindex
            Geometries for roads in gdb file, in exclusion coordinate
            system
        """
        lyr = fiona.listlayers(roads_fpath)[0]
        roads = gpd.read_file(roads_fpath, driver='FileGDB', layer=lyr)

        return roads.to_crs(crs=crs)

    @staticmethod
    def _get_feature_paths(roads_path):
        """
        Find all roads gdb files in roads_dir

        Parameters
        ----------
        roads_path : str
            Path to state here streets gdb file or directory containing
            states gdb files. Used to identify roads to build setbacks from.
            Files should be by state

        Returns
        -------
        file_paths : list
            List of file paths to all roads .gdp files in roads_dir
        """
        if roads_path.endswith('.gdb'):
            file_paths = [roads_path]
        else:
            file_paths = []
            for file in sorted(os.listdir(roads_path)):
                if file.endswith('.gdb') and file.startswith('Streets_USA'):
                    file_paths.append(os.path.join(roads_path, file))

        return file_paths

    def _parse_regs(self, regs_fpath):
        """
        Parse wind regulations, reduce table to just roads

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file

        Returns
        -------
        regs : pandas.DataFrame
            Wind regulations table
        """
        regs = super()._parse_regs(regs_fpath)

        mask = regs['Feature Type'].isin(['Roads', 'Highways', 'Highways 111'])
        regs = regs.loc[mask]

        return regs

    def _check_regs(self, features_fpath):
        """
        Reduce regs to state corresponding to features_fpath if needed

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from

        Returns
        -------
        wind_regs : geopands.GeoDataFrame | None
            Wind Regulations
        """
        state = features_fpath.split('.')[0].split('_')[-1]
        wind_regs = self.wind_regs
        mask = wind_regs['Abbr'] == state

        if not mask.any():
            msg = ("There are no local regulations in {}!".format(state))
            logger.error(msg)
            raise RuntimeError(msg)

        wind_regs = wind_regs.loc[mask].reset_index(drop=True)
        logger.debug('Computing setbacks for wind regulations in {} counties'
                     .format(len(wind_regs)))

        return wind_regs

    @classmethod
    def run(cls, excl_h5, roads_path, out_dir, hub_height,
            rotor_diameter, regs_fpath=None, multiplier=None,
            chunks=(128, 128), max_workers=None, replace=False, hsds=False):
        """
        Compute state's road setbacks and write them to a geotiff.
        If a wind regulations file is given compute local setbacks, otherwise
        compute generic setbacks using the given multiplier and the turbine
        tip-height.

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers
        road_path : str
            Path to state here streets gdb file or directory containing
            states gdb files.
        out_dir : str
            Directory to save setbacks geotiff(s) into
        hub_height : float | int
            Turbine hub height (m), used along with rotor diameter to compute
            blade tip height which is used to determine setback distance
        rotor_diameter : float | int
            Turbine rotor diameter (m), used along with hub height to compute
            blade tip height which is used to determine setback distance
        regs_fpath : str | None, optional
            Path to wind regulations .csv file, if None create generic
            setbacks using max-tip height * "multiplier", by default None
        multiplier : int | float | str | None, optional
            setback multiplier to use if wind regulations are not supplied,
            if str, must a key in {'high': 3, 'moderate': 1.1}, if supplied
            along with regs_fpath, will be ignored, multiplied with
            max-tip height, by default None
        chunks : tuple, optional
            Chunk size to use for setback layers, if None use default chunk
            size in excl_h5, by default (128, 128)
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None
        replace : bool, optional
            Flag to replace geotiff if it already exists, by default False
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        """
        setbacks = cls(excl_h5, hub_height, rotor_diameter,
                       regs_fpath=regs_fpath, multiplier=multiplier,
                       hsds=hsds, chunks=chunks)

        roads_path = setbacks._get_feature_paths(roads_path)
        for fpath in roads_path:
            geotiff = os.path.basename(fpath).replace('.gdb', '.tif')
            geotiff = os.path.join(out_dir, geotiff)
            if os.path.exists(geotiff) and not replace:
                msg = ('{} already exists, setbacks will not be re-computed '
                       'unless replace=True'.format(geotiff))
                logger.error(msg)
            else:
                logger.info("Computing setbacks from roads in {} and saving "
                            "to {}".format(fpath, geotiff))
                setbacks.compute_setbacks(fpath, geotiff=geotiff,
                                          max_workers=max_workers,
                                          replace=replace)


class TransmissionWindSetbacks(BaseWindSetbacks):
    """
    Transmission Wind setbacks, computed against a single set of transmission
    features instead of against state level features
    """
    @staticmethod
    def _compute_local_setbacks(features, cnty, setback):
        """
        Compute local county setbacks

        Parameters
        ----------
        features : geopandas.GeoDataFrame
            Features to setback from
        cnty : geopandas.GeoDataFrame
            Wind regulations for a single county
        setback : int
            Setback distance in meters

        Returns
        -------
        setbacks : list
            List of setback geometries
        """
        logger.debug('- Computing setbacks for county FIPS {}'
                     .format(cnty.iloc[0]['FIPS']))
        log_mem(logger)
        tmp = gpd.clip(features, cnty)
        tmp = tmp[~tmp.is_empty]

        # Buffer setback
        tmp.loc[:, 'geometry'] = tmp.buffer(setback)

        setbacks = [(geom, 1) for geom in tmp['geometry']]

        return setbacks

    @staticmethod
    def _get_feature_paths(features_path):
        """
        Ensure features path is valid

        Parameters
        ----------
        features_path : str
            Path to features file

        Returns
        -------
        features_path : list
            Features path as a list
        """
        if not os.path.exists(features_path):
            msg = '{} is not a valid file path!'.format(features_path)
            logger.error(msg)
            raise FileNotFoundError(msg)

        return [features_path]

    def _parse_regs(self, regs_fpath):
        """
        Parse wind regulations, reduce table to just transmission

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file

        Returns
        -------
        regs : pandas.DataFrame
            Wind regulations table
        """
        regs = super()._parse_regs(regs_fpath)

        mask = regs['Feature Type'] == 'Transmission'
        regs = regs.loc[mask]

        return regs

    @classmethod
    def run(cls, excl_h5, features_fpath, out_dir, hub_height,
            rotor_diameter, regs_fpath=None, multiplier=None,
            chunks=(128, 128), max_workers=None, replace=False, hsds=False):
        """
        Compute setbacks from given features and write them to a geotiff.
        If a wind regulations file is given compute local setbacks, otherwise
        compute generic setbacks using the given multiplier and the turbine
        tip-height.
        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers
        features_fpath : str
            Path to shape file with transmission or rail features to compute
            setbacks from
        out_dir : str
            Directory to save geotiff containing rasterized setbacks into
        hub_height : float | int
            Turbine hub height (m), used along with rotor diameter to compute
            blade tip height which is used to determine setback distance
        rotor_diameter : float | int
            Turbine rotor diameter (m), used along with hub height to compute
            blade tip height which is used to determine setback distance
        regs_fpath : str | None, optional
            Path to wind regulations .csv file, if None create generic
            setbacks using max-tip height * "multiplier", by default None
        multiplier : int | float | str | None, optional
            setback multiplier to use if wind regulations are not supplied,
            if str, must a key in {'high': 3, 'moderate': 1.1}, if supplied
            along with regs_fpath, will be ignored, multiplied with
            max-tip height, by default None
        chunks : tuple, optional
            Chunk size to use for setback layers, if None use default chunk
            size in excl_h5, by default (128, 128)
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None
        replace : bool, optional
            Flag to replace geotiff if it already exists, by default False
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        """
        geotiff = os.path.basename(features_fpath).split('.')[0]
        geotiff += '.tif'
        geotiff = os.path.join(out_dir, geotiff)
        if os.path.exists(geotiff) and not replace:
            msg = ('{} already exists, setbacks will not be re-computed '
                   'unless replace=True'.format(geotiff))
            logger.error(msg)
        else:
            setbacks = cls(excl_h5, hub_height, rotor_diameter,
                           regs_fpath=regs_fpath, multiplier=multiplier,
                           hsds=hsds, chunks=chunks)

            logger.info("Computing setbacks from {} and saving "
                        "to {}".format(features_fpath, geotiff))
            setbacks.compute_setbacks(features_fpath, geotiff=geotiff,
                                      max_workers=max_workers,
                                      replace=replace)


class RailWindSetbacks(TransmissionWindSetbacks):
    """
    Rail Wind setbacks, computed against a single set of railroad features,
    instead of state level features, uses the same approach as
    TransmissionWindSetbacks
    """
    def _parse_regs(self, regs_fpath):
        """
        Parse wind regulations, reduce table to just rail

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file

        Returns
        -------
        regs : pandas.DataFrame
            Wind regulations table
        """
        # pylint: disable=bad-super-call
        regs = super(TransmissionWindSetbacks, self)._parse_regs(regs_fpath)

        mask = regs['Feature Type'] == 'Railroads'
        regs = regs.loc[mask]

        return regs



# -*- coding: utf-8 -*-
"""
Module to compute hybrid solar-wind generation correlations
"""
from concurrent.futures import as_completed
import logging
import numpy as np
import os
import pandas as pd
from scipy.spatial import cKDTree
from scipy.stats import pearsonr, spearmanr, kendalltau

from rex.resource import Resource
from rex.utilities.utilities import parse_year
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.utilities import get_lat_lon_cols, roll_timeseries

logger = logging.getLogger(__name__)


def pearson_correlation(solar_ts, wind_ts):
    """
    Compute the Pearson correlation coefficient between the given solar
    and wind timeseries data. Return just the correlation coefficient.

    Parameters
    ----------
    solar_ts : ndarray
        Solar time-series vector for a single site
    wind_ts : ndarray
        Wind time-series vector for a single site

    Returns
    -------
    float
        Pearson correlation coefficient
    """
    return pearsonr(solar_ts, wind_ts)[0]


def spearman_correlation(solar_ts, wind_ts):
    """
    Compute the Spearman correlation coefficient between the given solar
    and wind timeseries data. Return just the correlation coefficient.

    Parameters
    ----------
    solar_ts : ndarray
        Solar time-series vector for a single site
    wind_ts : ndarray
        Wind time-series vector for a single site

    Returns
    -------
    float
        Spearman correlation coefficient
    """
    return spearmanr(solar_ts, wind_ts)[0]


def kendall_tau(solar_ts, wind_ts):
    """
    Compute Kendall's tau correlation between the given solar
    and wind timeseries data. Return just the correlation coefficient.

    Parameters
    ----------
    solar_ts : ndarray
        Solar time-series vector for a single site
    wind_ts : ndarray
        Wind time-series vector for a single site

    Returns
    -------
    float
        Kendall's tau
    """
    return kendalltau(solar_ts, wind_ts)[0]


class HybridStats:
    """
    Compute comparitive correlations for co-located wind and solar generation
    """
    STATS = {'pearson': {'func': pearson_correlation},
             'spearman': {'func': spearman_correlation},
             'kendall': {'func': kendall_tau}}

    def __init__(self, solar_h5, wind_h5, statistics='pearson',
                 res_cls=Resource, year=None):
        """
        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        statistics : str | tuple | dict, optional
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}},
            by default 'pearson'
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None
        """
        self._solar_h5 = solar_h5
        self._wind_h5 = wind_h5
        self._stats = None
        self.statistics = statistics

        self._res_cls = res_cls
        out = self._pre_flight_check(year=year)
        self._meta, self._time_index = out[:2]
        self._solar_time_slice, self._wind_time_slice = out[2:]

    def __repr__(self):
        msg = ('Computing {} between {} and {}'
               .format(list(self.statistics), self.solar_h5, self.wind_h5))

        return msg

    @property
    def solar_h5(self):
        """
        Path to solar h5 file(s)

        Returns
        -------
        str
        """
        return self._solar_h5

    @property
    def wind_h5(self):
        """
        Path to wind h5 file(s)

        Returns
        -------
        str
        """
        return self._wind_h5

    @property
    def statistics(self):
        """
        Dictionary of statistic functions/kwargs to run

        Returns
        -------
        dict
        """
        return self._stats

    @statistics.setter
    def statistics(self, statistics):
        """
         Statistics to extract, either a key or tuple of keys in
        cls.STATS, or a dictionary of the form
        {'stat_name': {'func': *, 'kwargs: {**}}}

        Parameters
        ----------
        statistics : dict
        """
        if isinstance(statistics, str):
            statistics = (statistics, )

        if isinstance(statistics, (tuple, list)):
            statistics = {s: self.STATS[s] for s in statistics}

        for stat in statistics.values():
            msg = 'A "func"(tion) must be provided for each statistic'
            assert 'func' in stat, msg
            if 'kwargs' in stat:
                msg = 'statistic function kwargs must be a dictionary '
                assert isinstance(stat['kwargs'], dict), msg

        self._stats = statistics

    @property
    def res_cls(self):
        """
        Resource class to use to access res_h5

        Returns
        -------
        Class
        """
        return self._res_cls

    @property
    def time_index(self):
        """
        Resource Datetimes

        Returns
        -------
        pandas.DatetimeIndex
        """
        return self._time_index

    @property
    def meta(self):
        """
        Resource meta-data table

        Returns
        -------
        pandas.DataFrame
        """
        return self._meta

    @property
    def lat_lon(self):
        """
        Resource (lat, lon) coordinates

        Returns
        -------
        pandas.DataFrame
        """
        lat_lon_cols = get_lat_lon_cols(self.meta)

        return self.meta[lat_lon_cols]

    @staticmethod
    def _map_sites(solar_meta, wind_meta):
        """
        Map solar to wind sites retaining the higher resolution meta data.
        Mapping it done using a cKDTree to determine the closest coarser
        resolution site to each finer resolution site.

        Parameters
        ----------
        solar_meta : pandas.DataFrame
            Solar sites meta data table
        wind_meta : pandas.DataFrame
            Wind sites meta data table

        Returns
        -------
        meta : pandas.DataFrame
            Meta data table mapping the finer resolution data to the coarser
        """
        solar_n = len(solar_meta)
        solar_cols = get_lat_lon_cols(solar_meta)
        wind_n = len(wind_meta)
        wind_cols = get_lat_lon_cols(wind_meta)

        # pylint: disable=not-callable
        if solar_n < wind_n:
            tree = cKDTree(solar_meta[solar_cols].values)
            meta = wind_meta.rename(columns={'cap': 'wind_cap'})
            meta.index.name = 'wind_gid'
            meta = meta.reset_index()

            dist, pos = tree.query(meta[wind_cols].values)
            mask = dist <= np.median(dist) * 1.5
            pos = pos[mask]
            meta = meta.loc[mask]
            meta['solar_gid'] = pos
            if 'capacity' in solar_meta:
                meta['solar_cap'] = solar_meta['capacity'].values[pos]
        else:
            tree = cKDTree(wind_meta[wind_cols].values)
            meta = solar_meta.rename(columns={'cap': 'solar_cap'})
            meta.index.name = 'solar_gid'
            meta = meta.reset_index()

            dist, pos = tree.query(meta[solar_cols])
            mask = dist <= np.median(dist) * 1.5
            meta = meta.loc[mask]
            pos = pos[mask]
            meta['wind_gid'] = pos
            if 'capacity' in wind_meta:
                meta['wind_cap'] = wind_meta['capacity'].values[pos]

        return meta

    @staticmethod
    def _check_time_index(solar_time_index, wind_time_index):
        """
        Compare solar and wind time indexes. Determine coincident time_index
        and requisite time_slices to reduce solar and wind dataset coincident

        Parameters
        ----------
        solar_time_index : pandas.DatatimeIndex
            Datetime index for solar data, respresents the temporal resolution
            of the solar datasets
        wind_time_index : pands.DatatimeINdex
            Datetime index for wind data, respresents the temporal resolution
            of the wind datasets

        Returns
        -------
        time_index : pandas.DatatimeIndex
            Coincident datetime index between solar and wind datasets,
            i.e. datetime steps that are in both wind and solar data
        solar_time_slice : slice | ndarray
            slice or boolean index of the solar timesteps that are in the
            coincident time_index
        wind_time_slice : slice | ndarray
            slice or boolean index of the wind timesteps that are in the
            coincident time_index
        """
        solar_n = len(solar_time_index)
        solar_time_slice = slice(None)
        wind_n = len(wind_time_index)
        wind_time_slice = slice(None)
        if solar_n < wind_n:
            wind_time_slice = wind_time_index.isin(solar_time_index)
            time_index = solar_time_index
            if not wind_time_slice.all():
                msg = ("The following timesteps are not in both wind and "
                       "solar dataset:\n{}"
                       .format(wind_time_index[wind_time_slice]))
                logger.error(msg)
                raise RuntimeError(msg)
        elif wind_n < solar_n:
            solar_time_slice = solar_time_index.isin(wind_time_index)
            time_index = wind_time_slice
            if not solar_time_slice.all():
                msg = ("The following timesteps are not in both wind and "
                       "solar dataset:\n{}"
                       .format(solar_time_index[solar_time_slice]))
                logger.error(msg)
                raise RuntimeError(msg)
        else:
            time_index = solar_time_index

        return time_index, solar_time_slice, wind_time_slice

    @staticmethod
    def _groupby_data(data, diurnal=False, doy=False, month=False):
        """
        Groupby data by month and/or hour

        Parameters
        ----------
        data : pandas.DataFrame
            DataFrame of data where index is time_index, columns are sites
        diurnal : bool, optional
            Extract diurnal stats, by default False
        doy : bool, optional
            Extract doy-of-year stats, by default False
        month : bool, optional
            Extract monthly stats, by default False

        Returns
        -------
        data : pandas.Groupby
            Input DataFrame grouped by month and or hour if requested
        """
        groupby = [data.index.year]
        if month:
            groupby.append(data.index.month)

        if doy:
            groupby.append(data.index.dayofyear)

        if diurnal:
            groupby.append(data.index.hour)

        data = data.groupby(groupby)

        return data

    @staticmethod
    def _format_grp_names(grp_names):
        """
        Format groupby index values

        Parameters
        ----------
        grp_names : list
            Group by index values, these correspond to each unique group in
            the groupby

        Returns
        -------
        out : ndarray
            2D array of grp index values properly formatted as strings
        """
        month_map = {1: 'Jan', 2: 'Feb', 3: 'Mar', 4: 'Apr', 5: 'May',
                     6: 'June', 7: 'July', 8: 'Aug', 9: 'Sept', 10: 'Oct',
                     11: 'Nov', 12: 'Dec'}

        # pylint: disable=unnecessary-lambda
        year = lambda s: "{}".format(s)
        month = lambda s: "{}".format(month_map[s])
        hour = lambda s: "{:02d}:00UTC".format(s)

        grp_names = np.array(grp_names).T
        if len(grp_names.shape) == 1:
            grp_names = np.expand_dims(grp_names, 0)

        out = []
        for grp_i in grp_names:  # pylint: disable=not-an-iterable
            grp_max = grp_i.max()
            if grp_max <= 12:
                out.append(list(map(month, grp_i)))
            elif grp_max <= 23:
                out.append(list(map(hour, grp_i)))
            else:
                out.append(list(map(year, grp_i)))

        return np.array(out).T

    @classmethod
    def _create_names(cls, index, stats):
        """
        Generate statistics names

        Parameters
        ----------
        index : pandas.Index | pandas.MultiIndex
            Temporal index, either month, hour, or (month, hour)
        stats : list
            Statistics to be computed

        Returns
        -------
        columns_map : dict
            Dictionary of column names to use for each statistic
        columns : list
            Column names to use
        """
        column_names = cls._format_grp_names(index)

        columns_map = {}
        columns = []
        for s in stats:
            cols = {i: '{}_{}'.format('-'.join(n), s) for i, n
                    in zip(index, column_names)}
            columns_map[s] = cols
            columns.extend(list(cols.values()))

        return columns_map, columns

    @classmethod
    def _compute_correlations(cls, solar_data, wind_data, statistics,
                              diurnal=False, doy=False, month=False):
        """
        Compute desired correlations for desired time intervals between
        solar and wind data

        Parameters
        ----------
        solar_data : pandas.DataFrame
            DataFrame of solar data. Index is time_index, columns are sites
        wind_data : pandas.DataFrame
            DataFrame of wind data. Index is time_index, columns are sites
        statistics : dict
            Dictionary of statistic functions/kwargs to run
        diurnal : bool, optional
            Extract diurnal stats, by default False
        doy : bool, optional
            Extract doy-of-year stats, by default False
        month : bool, optional
            Extract monthly stats, by default False

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        sites = solar_data.columns.values
        solar_data = cls._groupby_data(solar_data, diurnal=diurnal, doy=doy,
                                       month=month)
        wind_data = cls._groupby_data(wind_data, diurnal=diurnal, doy=doy,
                                      month=month)

        cols_map, col_names = cls._create_names(list(solar_data.groups),
                                                list(statistics))
        out_stats = pd.DataFrame(columns=col_names, index=sites,
                                 dtype=np.float32)
        for grp_name, solar_grp in solar_data:

            wind_grp = wind_data.get_group(grp_name)
            msg = ('solar and wind data shapes do not match! {} != {}'
                   .format(solar_grp.shape, wind_grp.shape))
            assert solar_grp.shape == wind_grp.shape, msg
            for i in sites:
                solar_ts = solar_grp.iloc[:, i]
                wind_ts = wind_grp.iloc[:, i]
                for name, stat in statistics.items():
                    col = cols_map[name][grp_name]
                    func = stat['func']
                    kwargs = stat.get('kwargs', {})
                    out_stats.at[i, col] = func(solar_ts, wind_ts, **kwargs)

        return out_stats

    @classmethod
    def _extract_stats(cls, solar_h5, wind_h5, dataset, sites,
                       solar_time_slice, wind_time_slice, statistics=None,
                       time_index=None, res_cls=Resource, diurnal=False,
                       doy=False, month=False, combinations=False):
        """
        Extract stats for given dataset, sites, and temporal extent

        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        dataset : tuple
            Datasets to compare, in the form: (solar_dataset, wind_dataset)
        sites : pandas.DataFrame
            Subset of meta DataFrame with sites to extract
        solar_time_slice : slice | ndarray
            slice or boolean index of the solar timesteps that are in the
            coincident time_index
        wind_time_slice : slice | ndarray
            slice or boolean index of the wind timesteps that are in the
            coincident time_index
        statistics : dict, optional
            Dictionary of statistic functions/kwargs to run, if None default
            to: {'pearson', {'func': pearson_correlation}},
            by default None
        time_index : pandas.DatatimeIndex, optional
            Timeseries DatetimeIndex, if None extract from wind_h5,
            by default None
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        diurnal : bool, optional
            Extract diurnal stats, by default False
        doy : bool, optional
            Extract doy-of-year stats, by default False
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        if statistics is None:
            statistics = {'pearson': {'func': pearson_correlation}}

        wind_sites = sites['wind_gid'].values
        solar_sites = sites['solar_gid'].values

        solar_dataset, wind_dataset = dataset
        with res_cls(wind_h5) as f:
            if time_index is None:
                time_index = f.time_index

            wind_data = f[wind_dataset, wind_time_slice, wind_sites]
            wind_data = pd.DataFrame(wind_data, index=time_index)

        with res_cls(solar_h5) as f:
            solar_data = f[solar_dataset, solar_time_slice, solar_sites]
            solar_data = pd.DataFrame(solar_data, index=time_index)

        if combinations:
            out_stats = [cls._compute_correlations(solar_data, wind_data,
                                                   statistics)]
            if month:
                out_stats.append(cls._compute_correlations(solar_data,
                                                           wind_data,
                                                           statistics,
                                                           month=True))

            if doy:
                out_stats.append(cls._compute_correlations(solar_data,
                                                           wind_data,
                                                           statistics,
                                                           doy=True))

            if diurnal:
                out_stats.append(cls._compute_correlations(solar_data,
                                                           wind_data,
                                                           statistics,
                                                           diurnal=True))
            if month and diurnal:
                out_stats.append(cls._compute_correlations(solar_data,
                                                           wind_data,
                                                           statistics,
                                                           month=True,
                                                           diurnal=True))

            out_stats = pd.concat(out_stats, axis=1)
        else:
            out_stats = cls._compute_correlations(solar_data,
                                                  wind_data,
                                                  statistics,
                                                  diurnal=diurnal,
                                                  doy=doy,
                                                  month=month)

        out_stats.index = sites.index.values

        out_stats.index.name = 'gid'

        return out_stats

    @staticmethod
    def save_stats(out_stats, out_fpath):
        """
        Save correlations to disk

        Parameters
        ----------
        out_stats : pandas.DataFrame
            Table of correlations to save
        out_path : str
            .csv, or .json path to save statistics too
        """
        logger.info('Saving hybrid stats to {}'.format(out_fpath))
        if out_fpath.endswith('.csv'):
            out_stats.to_csv(out_fpath)
        elif out_fpath.endswith('.json'):
            out_stats.to_json(out_fpath)
        else:
            msg = ("Cannot save statistics, expecting a .csv, or "
                   ".json path, but got: {}".format(out_fpath))
            logger.error(msg)
            raise OSError(msg)

    @staticmethod
    def _check_dataset(dataset):
        """
        Check user provided dataset for proper format

        Parameters
        ----------
        dataset : tuple | str
            Dataset to compare, if a string, extract the same
            dataset for both with and solar, other wise a tuple of the form:
            (solar_dataset, wind_dataset)

        Returns
        -------
        dataset : tuple
            Datasets to compare, in the form: (solar_dataset, wind_dataset)
        """
        if isinstance(dataset, str):
            dataset = (dataset, dataset)
        elif isinstance(dataset, (tuple, list)):
            if len(dataset) < 2:
                msg = ("Must supply a solar and wind dataset in the form: "
                       "(solar, wind)")
                logger.error(msg)
                raise ValueError(msg)

        return dataset

    @staticmethod
    def _parse_meta_time_index(h5_path, res_cls=Resource, year=None):
        """
        Parse meta data table and time_index from .h5 file. If 'year' is
        provided extact time_index for given year.

        Parameters
        ----------
        h5_path : str
            Path to .h5 file to extract meta and time_index
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None

        Returns
        -------
        meta : pandas.DataFrame
            Site meta data table
        time_index: pandas.DatatimeIndex
            Datetime Index
        """
        with res_cls(h5_path) as f:
            meta = f.meta
            if 'time_index' in f:
                time_index = f.time_index
            elif year is not None:
                time_index = f[f'time_index-{year}']
            else:
                ti_dsets = [dset for dset in f.datasets
                            if dset.startswith('time_index')]
                msg = ("'time_index' is not available in {}. The following "
                       "potential annual time_index are available: {}. Please "
                       "specify a 'year' to use the 'time_index' for a "
                       "specific year.".format(h5_path, ti_dsets))
                logger.error(ti_dsets)
                raise ValueError(msg)

        return meta, time_index

    def _pre_flight_check(self, year=None):
        """
        Compare solar and wind site meta data and time index to ensure
        they can be compared

        Parameters
        ----------
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None

        Returns
        -------
        meta : pandas.DataFrame
            Meta data table mapping the finer resolution data to the coarser
        time_index : pandas.DatatimeIndex
            Coincident datetime index between solar and wind datasets,
            i.e. datetime steps that are in both wind and solar data
        solar_time_slice : slice | ndarray
            slice or boolean index of the solar timesteps that are in the
            coincident time_index
        wind_time_slice : slice | ndarray
            slice or boolean index of the wind timesteps that are in the
            coincident time_index
        """
        solar_meta, solar_ti = self._parse_meta_time_index(
            self._solar_h5, res_cls=self._res_cls, year=year)
        wind_meta, wind_ti = self._parse_meta_time_index(
            self._wind_h5, res_cls=self._res_cls, year=year)

        time_index, solar_time_slice, wind_time_slice = \
            self._check_time_index(solar_ti, wind_ti)
        meta = self._map_sites(solar_meta, wind_meta)

        return meta, time_index, solar_time_slice, wind_time_slice

    def _compute_stats(self, dataset, max_workers=None, sites_per_worker=1000,
                       lat_lon_only=True, extract_stats_kwargs=None):
        """
        Compute correlations

        Parameters
        ----------
        dataset : tuple | str
            Dataset to compare, if a string, extract the same
            dataset for both with and solar, other wise a tuple of the form:
            (solar_dataset, wind_dataset)
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        sites_per_worker : int, optional
            Number of sites to extract on each worker, by default 1000
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        extract_stats_kwargs : dict, optional
            Kwargs to pass to _extract_stats method, by default None

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of desired correlation coefficients at desired time
            intervals
        """
        if extract_stats_kwargs is None:
            extract_stats_kwargs = {'res_cls': self.res_cls}

        dataset = self._check_dataset(dataset)

        solar_h5 = self.solar_h5
        wind_h5 = self.wind_h5
        if max_workers is None:
            max_workers = os.cpu_count()

        slices = len(self.meta) // sites_per_worker
        if slices:
            slices = np.array_split(self.meta, slices)
        else:
            slices = [self.meta]
            max_workers = 1

        if max_workers > 1:
            msg = ('Extracting {} for {} in parallel using {} workers'
                   .format(self, dataset, max_workers))
            logger.info(msg)

            loggers = [__name__, 'reVX', 'rex']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for sites in slices:
                    future = exe.submit(self._extract_stats,
                                        solar_h5, wind_h5, dataset, sites,
                                        self._solar_time_slice,
                                        self._wind_time_slice,
                                        **extract_stats_kwargs)
                    futures.append(future)

                out_stats = []
                for i, future in enumerate(as_completed(futures)):
                    out_stats.append(future.result())
                    logger.debug('Completed {} out of {} workers'
                                 .format((i + 1), len(futures)))

            out_stats = pd.concat(out_stats)
        else:
            msg = ('Extracting {} for {} in serial'
                   .format(self, dataset))
            logger.info(msg)
            out_stats = []
            for i, sites in enumerate(slices):
                out_stats.append(self._extract_stats(
                    solar_h5, wind_h5, dataset, sites,
                    self._solar_time_slice,
                    self._wind_time_slice,
                    **extract_stats_kwargs))
                logger.debug('Completed {} out of {} sets of sites'
                             .format((i + 1), len(slices)))

            out_stats = pd.concat(out_stats)

        if lat_lon_only:
            meta = self.lat_lon
        else:
            meta = self.meta

        out_stats = meta.join(out_stats.sort_index(), how='inner')

        return out_stats

    def compute_stats(self, dataset, diurnal=False, doy=False, month=False,
                      combinations=False, max_workers=None,
                      sites_per_worker=1000, lat_lon_only=True):
        """
        Compute correlations

        Parameters
        ----------
        dataset : tuple | str
            Dataset to compare, if a string, extract the same
            dataset for both with and solar, other wise a tuple of the form:
            (solar_dataset, wind_dataset)
        diurnal : bool, optional
            Extract diurnal stats, by default False
        doy : bool, optional
            Extract doy-of-year stats, by default False
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        sites_per_worker : int, optional
            Number of sites to extract on each worker, by default 1000
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True

        Returns
        -------
        res_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        kwargs = {'res_cls': self.res_cls,
                  'diurnal': diurnal,
                  'doy': doy,
                  'month': month,
                  'combinations': combinations,
                  'statistics': self.statistics,
                  'time_index': self.time_index
                  }

        logger.info('Computing correlations from {}'.format(dataset))
        logger.debug('- Using the following options: {}'.format(kwargs))
        out_stats = self._compute_stats(dataset, max_workers=max_workers,
                                        sites_per_worker=sites_per_worker,
                                        lat_lon_only=lat_lon_only,
                                        extract_stats_kwargs=kwargs)

        return out_stats

    @classmethod
    def run(cls, solar_h5, wind_h5, dataset, statistics='pearson',
            diurnal=False, doy=False, month=False, combinations=False,
            res_cls=Resource, year=None, max_workers=None,
            sites_per_worker=1000, lat_lon_only=True, out_path=None):
        """
        Compute temporal stats between solar and wind time-series at desired
        temporal scales

        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        dataset : tuple | str
            Dataset to compare, if a string, extract the same
            dataset for both with and solar, other wise a tuple of the form:
            (solar_dataset, wind_dataset)
        statistics : str | tuple | dict, optional
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}},
            by default 'pearson'
        diurnal : bool, optional
            Extract diurnal stats, by default False
        doy : bool, optional
            Extract doy-of-year stats, by default False
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        sites_per_worker : int, optional
            Number of sites to extract on each worker, by default 1000
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            .csv, or .json path to save statistics too, by default None

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of resource statistics
        """
        if isinstance(dataset, str) and year is None:
            try:
                year = parse_year(dataset)
            except RuntimeError:
                year = None

        hybrid_stats = cls(solar_h5, wind_h5, statistics=statistics,
                           res_cls=res_cls, year=year)
        out_stats = hybrid_stats.compute_stats(
            dataset, diurnal=diurnal, doy=doy, month=month,
            combinations=combinations, max_workers=max_workers,
            sites_per_worker=sites_per_worker, lat_lon_only=lat_lon_only)
        if out_path is not None:
            hybrid_stats.save_stats(out_stats, out_path)

        return out_stats

    @classmethod
    def cf_profile(cls, solar_h5, wind_h5, statistics='pearson',
                   diurnal=False, doy=False, month=False, combinations=False,
                   res_cls=Resource, max_workers=None,
                   sites_per_worker=1000, lat_lon_only=True, out_path=None):
        """
        Compute temporal stats on cf_profile dataset

        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        statistics : str | tuple | dict, optional
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}},
            by default 'pearson'
        diurnal : bool, optional
            Extract diurnal stats, by default False
        doy : bool, optional
            Extract doy-of-year stats, by default False
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        sites_per_worker : int, optional
            Number of sites to extract on each worker, by default 1000
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            .csv, or .json path to save statistics too, by default None

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of resource statistics
        """
        out_stats = cls.run(solar_h5, wind_h5, 'cf_profile',
                            statistics=statistics, res_cls=res_cls,
                            diurnal=diurnal, doy=doy, month=month,
                            combinations=combinations, max_workers=max_workers,
                            sites_per_worker=sites_per_worker,
                            lat_lon_only=lat_lon_only, out_path=out_path)

        return out_stats


class HybridCrossCorrelation(HybridStats):
    """
    Compute the temporal cross correlations for co-located wind and solar
    generation
    """

    def __init__(self, solar_h5, wind_h5, res_cls=Resource, year=None):
        """
        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None
        """
        self._solar_h5 = solar_h5
        self._wind_h5 = wind_h5
        self._res_cls = res_cls
        self._stats = None

        out = self._pre_flight_check(year=year)
        self._meta, self._time_index = out[:2]
        self._solar_time_slice, self._wind_time_slice = out[2:]

    def __repr__(self):
        msg = ('Computing cross-correlations between {} and {}'
               .format(self.solar_h5, self.wind_h5))

        return msg

    @staticmethod
    def cross_correlation(solar_data, wind_data, m):
        """
        Compute the cross-correlation between solar and wind time-series data
        with time-lag m

        Parameters
        ----------
        solar_data : ndarray
            Time-series solar data
        wind_data : ndarray
            Time-series wind data
        m : int
            Integer shift between solar and wind time-series, is pass directly
            to np.roll, so it will be the number of time-steps that are
            shifted. The lag time will be m * dt where dt is the time-step
            size.

        Returns
        -------
        corr : ndarray
            Cross-correlation coefficient for each solar, wind site pair
        """
        solar_u = solar_data.mean(axis=0)
        solar_s = solar_data.std(axis=0)
        solar_data = np.abs((solar_data - solar_u) / solar_s)

        wind_u = wind_data.mean(axis=0)
        wind_s = wind_data.mean(axis=0)
        wind_data = np.roll(wind_data, m, axis=0)
        wind_data = np.abs((wind_data - wind_u) / wind_s)

        n = len(solar_data)
        corr = (1 / (n - 1)) * np.sum(solar_data * wind_data, axis=0)

        return corr

    @classmethod
    def _extract_stats(cls, solar_h5, wind_h5, dataset, sites,
                       solar_time_slice, wind_time_slice,
                       max_lag=50, lag_step=1, res_cls=Resource):
        """
        Extract stats for given dataset, sites, and temporal extent

        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        dataset : tuple
            Datasets to compare, in the form: (solar_dataset, wind_dataset)
        sites : pandas.DataFrame
            Subset of meta DataFrame with sites to extract
        solar_time_slice : slice | ndarray
            slice or boolean index of the solar timesteps that are in the
            coincident time_index
        wind_time_slice : slice | ndarray
            slice or boolean index of the wind timesteps that are in the
            coincident time_index
        max_lag : int, optional
            Maximum lag size. Cross-correlation will be run for all lags in
            range(-max_lag, max_lag + lag_step, lag_step), each value in the
            range is the number of timesteps by which the time-series will be
            shifted to compute the cross-correlation.
            by default 50
        lag_step : int, optional
            Step size used between lag sizes. Is the number of time-steps
            between lag sizes, by default 1
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        wind_sites = sites['wind_gid'].values
        solar_sites = sites['solar_gid'].values

        solar_dataset, wind_dataset = dataset
        with res_cls(solar_h5) as f:
            solar_data = f[solar_dataset, solar_time_slice, solar_sites]

        with res_cls(wind_h5) as f:
            wind_data = f[wind_dataset, wind_time_slice, wind_sites]

        out_stats = {}
        for m in range(-max_lag, max_lag + lag_step, lag_step):
            out_stats[m] = cls.cross_correlation(solar_data, wind_data, m)

        index = pd.Index(sites.index.values, name='gid')
        out_stats = pd.DataFrame(out_stats,
                                 index=index)

        out_stats['optimal_m'] = \
            out_stats.columns[out_stats.values.argmax(axis=1)]

        return out_stats

    def compute_stats(self, dataset, max_lag=50, lag_step=1,
                      max_workers=None, sites_per_worker=1000,
                      lat_lon_only=True):
        """
        Compute correlations

        Parameters
        ----------
        dataset : tuple | str
            Dataset to compare, if a string, extract the same
            dataset for both with and solar, other wise a tuple of the form:
            (solar_dataset, wind_dataset)
        max_lag : int, optional
            Maximum lag size. Cross-correlation will be run for all lags in
            range(-max_lag, max_lag, lag_step), by default 50
        lag_step : int, optional
            Step size used between lag sizes, by default 1
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        sites_per_worker : int, optional
            Number of sites to extract on each worker, by default 1000
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True

        Returns
        -------
        res_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        kwargs = {'res_cls': self.res_cls,
                  'max_lag': max_lag,
                  'lag_step': lag_step}

        logger.info('Computing cross correlations from {}'.format(dataset))
        logger.debug('- Using the following options: {}'.format(kwargs))
        out_stats = self._compute_stats(dataset, max_workers=max_workers,
                                        sites_per_worker=sites_per_worker,
                                        lat_lon_only=lat_lon_only,
                                        extract_stats_kwargs=kwargs)

        return out_stats

    @classmethod
    def run(cls, solar_h5, wind_h5, dataset, max_lag=50, lag_step=1,
            res_cls=Resource, year=None, max_workers=None,
            sites_per_worker=1000, lat_lon_only=True, out_path=None):
        """
        Compute cross correlations between solar and wind time-series

        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        dataset : tuple | str
            Dataset to compare, if a string, extract the same
            dataset for both with and solar, other wise a tuple of the form:
            (solar_dataset, wind_dataset)
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None
        max_lag : int, optional
            Maximum lag size. Cross-correlation will be run for all lags in
            range(-max_lag, max_lag, lag_step), by default 50
        lag_step : int, optional
            Step size used between lag sizes, by default 1
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        sites_per_worker : int, optional
            Number of sites to extract on each worker, by default 1000
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            .csv, or .json path to save statistics too, by default None

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of resource statistics
        """
        if isinstance(dataset, str) and year is None:
            try:
                year = parse_year(dataset)
            except RuntimeError:
                year = None

        hybrid_stats = cls(solar_h5, wind_h5, res_cls=res_cls, year=year)
        out_stats = hybrid_stats.compute_stats(
            dataset, max_lag=max_lag, lag_step=lag_step,
            max_workers=max_workers, sites_per_worker=sites_per_worker,
            lat_lon_only=lat_lon_only)
        if out_path is not None:
            hybrid_stats.save_stats(out_stats, out_path)

        return out_stats

    @classmethod
    def cf_profile(cls, solar_h5, wind_h5, max_lag=50, lag_step=1,
                   res_cls=Resource, max_workers=None,
                   sites_per_worker=1000, lat_lon_only=True, out_path=None):
        """
        Compute cross correlations on cf_profile dataset

        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        dataset : str
            Dataset to extract stats for
        max_lag : int, optional
            Maximum lag size. Cross-correlation will be run for all lags in
            range(-max_lag, max_lag, lag_step), by default 50
        lag_step : int, optional
            Step size used between lag sizes, by default 1
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        sites_per_worker : int, optional
            Number of sites to extract on each worker, by default 1000
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            .csv, or .json path to save statistics too, by default None

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of resource statistics
        """
        out_stats = cls.run(solar_h5, wind_h5, 'cf_profile', res_cls=res_cls,
                            max_lag=max_lag, lag_step=lag_step,
                            max_workers=max_workers,
                            sites_per_worker=sites_per_worker,
                            lat_lon_only=lat_lon_only, out_path=out_path)

        return out_stats


class HybridStabilityCoefficient(HybridStats):
    """
    Compute the annual/monthly stability coefficient for co-located wind and
    solar
    """

    def __init__(self, solar_h5, wind_h5, res_cls=Resource, year=None):
        """
        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None
        """
        self._solar_h5 = solar_h5
        self._wind_h5 = wind_h5
        self._res_cls = res_cls
        self._stats = None

        out = self._pre_flight_check(year=year)
        self._meta, self._time_index = out[:2]
        self._solar_time_slice, self._wind_time_slice = out[2:]

    def __repr__(self):
        msg = ('Computing stability coefficient between {} and {}'
               .format(self.solar_h5, self.wind_h5))

        return msg

    @staticmethod
    def _daily_variability(doy):
        """
        Compute the daily variability

        Parameters
        ----------
        doy : pandas.DataFrameGroupby
            Time-series DataFrame grouped by day-of-year

        Returns
        -------
        var : pandas.DataFrame
            Daily variablility by site
        """
        var = np.sqrt(np.sum((doy - doy.mean())**2))

        return var

    @classmethod
    def stability_coefficient(cls, mix, ref):
        """
        Compute average stability coefficient

        Parameters
        ----------
        mix : pandas.DataFrame
            DataFrame of mixed solar and wind time-series
        ref : pandas.DataFrame
            DataFrame of reference (solar or wind) time-series

        Returns
        -------
        stab : ndarray
            Vector of the average stability coefficient for all days in the
            provided time-series data. Averages are by site.
        """
        mix = mix.groupby(mix.index.dayofyear)
        mix_var = mix.apply(cls._daily_variability)

        ref = ref.groupby(ref.index.dayofyear)
        ref_var = ref.apply(cls._daily_variability)

        stab = 1 - ((mix_var / ref_var) * (ref.mean() / mix.mean()))

        mask = np.isfinite(stab)
        if not np.all(mask):
            stab[~mask] = np.nan

        return stab.mean().values.astype(np.float32)

    @classmethod
    def _compute_coefficients(cls, solar_data, wind_data, solar_cap=None,
                              wind_cap=None, month=False, reference='solar'):
        """
        Compute compute average stability coefficient of solar and wind data
        over desired time intervals

        Parameters
        ----------
        solar_data : pandas.DataFrame
            DataFrame of solar data. Index is time_index, columns are sites
        wind_data : pandas.DataFrame
            DataFrame of wind data. Index is time_index, columns are sites
        month : bool, optional
            Extract monthly stats, by default False
        reference : str, optional
            Which data to use as the reference (denominator) when computing
            the stability coefficient, by default 'solar'

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of stability coefficients for given sites and desired
            time intervals
        """
        sites = solar_data.columns.values
        if solar_cap is None or wind_cap is None:
            mix = (solar_data + wind_data) / 2
        else:
            mix = ((solar_data * solar_cap + wind_data * wind_cap)
                   / (solar_cap + wind_cap))

        mix = cls._groupby_data(mix, month=month)

        if reference.lower() == 'solar':
            ref = cls._groupby_data(solar_data, month=month)
        else:
            ref = cls._groupby_data(wind_data, month=month)

        cols_map, _ = cls._create_names(list(mix.groups),
                                        ['stability'])
        out_stats = {}
        for grp_name, mix_grp in mix:
            col = cols_map['stability'][grp_name]
            ref_grp = ref.get_group(grp_name)
            msg = ('mixed and reference data shapes do not match! {} != {}'
                   .format(mix_grp.shape, ref_grp.shape))
            assert mix_grp.shape == ref_grp.shape, msg
            out_stats[col] = cls.stability_coefficient(mix_grp, ref_grp)

        out_stats = out_stats = pd.DataFrame(out_stats, index=sites,
                                             dtype=np.float32)

        return out_stats

    @classmethod
    def _extract_stats(cls, solar_h5, wind_h5, dataset, sites,
                       solar_time_slice, wind_time_slice,
                       time_index=None, res_cls=Resource,
                       reference='solar', month=False,
                       combinations=False):
        """
        Extract stats for given dataset, sites, and temporal extent

        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        dataset : tuple
            Datasets to compare, in the form: (solar_dataset, wind_dataset)
        sites : pandas.DataFrame
            Subset of meta DataFrame with sites to extract
        solar_time_slice : slice | ndarray
            slice or boolean index of the solar timesteps that are in the
            coincident time_index
        wind_time_slice : slice | ndarray
            slice or boolean index of the wind timesteps that are in the
            coincident time_index
        time_index : pandas.DatatimeIndex, optional
            Timeseries DatetimeIndex, if None extract from wind_h5,
            by default None
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        reference : str, optional
            Which data to use as the reference (denominator) when computing
            the stability coefficient, by default 'solar'
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        solar_dataset, wind_dataset = dataset
        solar_sites = sites['solar_gid'].values
        tz = sites['timezone'].values.copy()

        solar_cap = None
        if 'solar_cap' in sites:
            solar_cap = sites['solar_cap'].values

        with res_cls(solar_h5) as f:
            solar_data = f[solar_dataset, solar_time_slice, solar_sites]
            solar_data = roll_timeseries(solar_data, tz)
            solar_data = pd.DataFrame(solar_data, index=time_index)

        wind_sites = sites['wind_gid'].values
        wind_cap = None
        if 'wind_cap' in sites:
            wind_cap = sites['wind_cap'].values

        with res_cls(wind_h5) as f:
            if time_index is None:
                time_index = f.time_index

            wind_data = f[wind_dataset, wind_time_slice, wind_sites]
            wind_data = roll_timeseries(wind_data, tz)
            wind_data = pd.DataFrame(wind_data, index=time_index)

        if combinations:
            out_stats = [cls._compute_coefficients(solar_data, wind_data,
                                                   solar_cap=solar_cap,
                                                   wind_cap=wind_cap,
                                                   reference=reference
                                                   )]
            if month:
                out_stats.append(cls._compute_coefficients(solar_data,
                                                           wind_data,
                                                           solar_cap=solar_cap,
                                                           wind_cap=wind_cap,
                                                           reference=reference,
                                                           month=True))

            out_stats = pd.concat(out_stats, axis=1)
        else:
            out_stats = cls._compute_coefficients(solar_data,
                                                  wind_data,
                                                  solar_cap=solar_cap,
                                                  wind_cap=wind_cap,
                                                  reference=reference,
                                                  month=month)

        out_stats.index = sites.index.values

        out_stats.index.name = 'gid'

        return out_stats

    def compute_stats(self, dataset, reference='solar', month=False,
                      combinations=False, max_workers=None,
                      sites_per_worker=1000, lat_lon_only=True):
        """
        Compute stability coefficients

        Parameters
        ----------
        dataset : tuple | str
            Dataset to compare, if a string, extract the same
            dataset for both with and solar, other wise a tuple of the form:
            (solar_dataset, wind_dataset)
        reference : str, optional
            Which data to use as the reference (denominator) when computing
            the stability coefficient, by default 'solar'
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        sites_per_worker : int, optional
            Number of sites to extract on each worker, by default 1000
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True

        Returns
        -------
        res_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        kwargs = {'time_index': self.time_index,
                  'res_cls': self.res_cls,
                  'month': month,
                  'combinations': combinations,
                  'reference': reference}

        logger.info('Computing stability coefficients from {}'.format(dataset))
        logger.debug('- Using the following options: {}'.format(kwargs))
        out_stats = self._compute_stats(dataset, max_workers=max_workers,
                                        sites_per_worker=sites_per_worker,
                                        lat_lon_only=lat_lon_only,
                                        extract_stats_kwargs=kwargs)

        return out_stats

    @classmethod
    def run(cls, solar_h5, wind_h5, dataset, reference='solar', month=False,
            combinations=False, res_cls=Resource, year=None, max_workers=None,
            sites_per_worker=1000, lat_lon_only=True, out_path=None):
        """
        Compute stability coefficient between solar and wind time-series.
        Time-series are shifted to local time before computing the daily
        stability coefficient. Final data is the average of daily stability
        coefficients for each month and/or year.

        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        dataset : tuple | str
            Dataset to compare, if a string, extract the same
            dataset for both with and solar, other wise a tuple of the form:
            (solar_dataset, wind_dataset)
        reference : str, optional
            Which data to use as the reference (denominator) when computing
            the stability coefficient, by default 'solar'
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        sites_per_worker : int, optional
            Number of sites to extract on each worker, by default 1000
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            .csv, or .json path to save statistics too, by default None

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of resource statistics
        """
        if isinstance(dataset, str) and year is None:
            try:
                year = parse_year(dataset)
            except RuntimeError:
                year = None

        hybrid_stats = cls(solar_h5, wind_h5, res_cls=res_cls, year=year)
        out_stats = hybrid_stats.compute_stats(
            dataset, month=month, combinations=combinations,
            reference=reference, max_workers=max_workers,
            sites_per_worker=sites_per_worker,
            lat_lon_only=lat_lon_only)
        if out_path is not None:
            hybrid_stats.save_stats(out_stats, out_path)

        return out_stats

    @classmethod
    def cf_profile(cls, solar_h5, wind_h5, reference='solar', month=False,
                   combinations=False, res_cls=Resource, max_workers=None,
                   sites_per_worker=1000, lat_lon_only=True, out_path=None):
        """
        Compute stability coefficient between solar and wind time-series.
        Time-series are shifted to local time before computing the daily
        stability coefficient. Final data is the average of daily stability
        coefficients for each month and/or year.

        Parameters
        ----------
        solar_h5 : str
            Path to solar h5 file(s)
        wind_h5 : str
            Path to wind h5 file(s)
        reference : str, optional
            Which data to use as the reference (denominator) when computing
            the stability coefficient, by default 'solar'
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        sites_per_worker : int, optional
            Number of sites to extract on each worker, by default 1000
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            .csv, or .json path to save statistics too, by default None

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of resource statistics
        """
        out_stats = cls.run(solar_h5, wind_h5, 'cf_profile',
                            reference=reference, month=month,
                            combinations=combinations, res_cls=res_cls,
                            max_workers=max_workers,
                            sites_per_worker=sites_per_worker,
                            lat_lon_only=lat_lon_only, out_path=out_path)

        return out_stats
